import numpy as np
from numba import njit

from src.Base.TicketToRide.docs.index import *


@njit()
def getAgentSize():
    return NUMBER_PLAYER


#########################################################


@njit()
def getActionSize():
    return NUMBER_ACTIONS


#########################################################


@njit()
def getStateSize():
    return P_LENGTH


#########################################################


@njit()
def initEnv():
    # tạo env_state full 0
    env_state = np.zeros(ENV_LENGTH)
    # các đường vô chủ gán là -1
    env_state[ENV_ROAD_BOARD:ENV_ROUTE_CARD_BOARD] = -1
    # các thẻ route đã bị người chơi sở hữu gán là -1
    env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER] = -1

    all_special_route_card = np.arange(
        NUMBER_ROUTE - NUMBER_SPECIAL_ROUTE, NUMBER_ROUTE
    )
    # chọn 5 thẻ special route cho người chơi
    special_route_card = np.random.choice(
        all_special_route_card, NUMBER_PLAYER, replace=False
    )

    all_normal_route_card = np.arange(NUMBER_ROUTE - NUMBER_SPECIAL_ROUTE)
    np.random.shuffle(all_normal_route_card)
    normal_route_card = all_normal_route_card[: NUMBER_PLAYER * NUMBER_ROUTE_GET]
    env_state[ENV_ROUTE_CARD_BOARD:ENV_TRAIN_CAR_CARD] = np.concatenate(
        (
            all_normal_route_card[NUMBER_PLAYER * NUMBER_ROUTE_GET :],
            np.full(NUMBER_SPECIAL_ROUTE + NUMBER_PLAYER * NUMBER_ROUTE_GET, -1),
        )
    )
    all_train_car_card = np.concatenate(
        (
            np.zeros(14),
            np.ones(12),
            np.full(12, 2),
            np.full(12, 3),
            np.full(12, 4),
            np.full(12, 5),
            np.full(12, 6),
            np.full(12, 7),
            np.full(12, 8),
        )
    )
    # các thẻ xe lửa
    np.random.shuffle(all_train_car_card)
    train_car_card = all_train_car_card[
        : NUMBER_PLAYER * NUMBER_TRAIN_CAR_GET
    ]  # các thẻ xe lửa cho người chơi
    # các thẻ xe lửa trong chồng bài úp
    env_state[
        ENV_TRAIN_CAR_CARD : ENV_TRAIN_CAR_CARD
        + NUMBER_TRAIN_CAR_CARD
        - NUMBER_PLAYER * NUMBER_TRAIN_CAR_GET
        - NUMBER_TRAIN_CAR_CARD_OPEN
    ] = all_train_car_card[
        NUMBER_PLAYER * NUMBER_TRAIN_CAR_GET + NUMBER_TRAIN_CAR_CARD_OPEN :
    ]

    for id_pl in range(NUMBER_PLAYER):
        special_route_id = special_route_card[id_pl]
        normal_route_id = normal_route_card[
            id_pl * NUMBER_ROUTE_GET : NUMBER_ROUTE_GET * (id_pl + 1)
        ]
        train_car_card_id = train_car_card[
            id_pl * NUMBER_TRAIN_CAR_GET : NUMBER_TRAIN_CAR_GET * (id_pl + 1)
        ]

        score_train_car = np.array(
            [0, 0, NUMBER_TRAIN]
        )  # điểm, điểm trừ, số toa tàu còn lại
        train_card_id = np.zeros(9)
        for id_train in train_car_card_id:
            train_card_id[int(id_train)] += 1
        route_card_id = np.zeros(NUMBER_ROUTE)
        route_card_id[special_route_id] = 1
        route_card_id[normal_route_id] = 1
        env_state[
            ENV_IN4_PLAYER
            + ATTRIBUTE_PLAYER * id_pl : ENV_IN4_PLAYER
            + ATTRIBUTE_PLAYER * (id_pl + 1)
        ] = np.concatenate((score_train_car, train_card_id, route_card_id))
    # các thẻ xe lửa được mở
    env_state[
        ENV_TRAIN_CAR_OPEN : ENV_TRAIN_CAR_OPEN + NUMBER_TRAIN_CAR_CARD_OPEN
    ] = all_train_car_card[
        NUMBER_PLAYER * NUMBER_TRAIN_CAR_GET : NUMBER_PLAYER * NUMBER_TRAIN_CAR_GET
        + NUMBER_TRAIN_CAR_CARD_OPEN
    ]
    env_state[ENV_ROUTE_CARD_GET : ENV_ROUTE_CARD_GET + NUMBER_ROUTE_RECEIVE] = -1
    # Other in4
    env_state[ENV_PHASE] = 3  # đầu game thì đi loại thẻ route
    env_state[ENV_CHECK_END] = 0
    env_state[ENV_ID_PLAYER_END] = -1
    env_state[ENV_ROAD_BUILT] = -1
    env_state[ENV_TURN] = 1
    id_action = int(env_state[ENV_ID_ACTION])
    env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP] = env_state[
        ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN
    ][id_action * ATTRIBUTE_PLAYER : (id_action + 1) * ATTRIBUTE_PLAYER][
        ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :
    ]
    env_state[ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN][
        id_action * ATTRIBUTE_PLAYER : (id_action + 1) * ATTRIBUTE_PLAYER
    ][ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :] = 0
    env_state[ENV_PHASE] = 3

    return env_state


#########################################################


@njit()
def getAgentState(env_state):
    player_state = np.zeros(P_LENGTH, dtype=np.float64)
    id_action = int(env_state[ENV_ID_ACTION])
    # điểm tất cả người chơi
    all_player_score = env_state[
        ENV_IN4_PLAYER + ATT_SCORE : ENV_TRAIN_CAR_OPEN : ATTRIBUTE_PLAYER
    ]

    player_state[P_SCORE:P_NEG_SCORE] = np.concatenate(
        (all_player_score[id_action:], all_player_score[:id_action])
    )
    # số thẻ train_car_card của người chơi
    player_state[P_TRAIN_CAR_CARD:P_PLAYER_ROAD] = env_state[
        ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN
    ][id_action * ATTRIBUTE_PLAYER : (id_action + 1) * ATTRIBUTE_PLAYER][
        ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
        + NUMBER_TYPE_TRAIN_CAR_CARD
    ]
    # số tàu người chơi còn
    player_state[P_NUMBER_TRAIN] = env_state[ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN][
        id_action * ATTRIBUTE_PLAYER + ATT_NUMBER_TRAIN
    ]
    # đường của các người chơi
    for id_ in range(NUMBER_PLAYER):
        list_road_id_ = np.where(env_state[ENV_ROAD_BOARD:ENV_ROUTE_CARD_BOARD] == id_)[
            0
        ]
        idx = (id_ - id_action) % NUMBER_PLAYER
        player_state[P_PLAYER_ROAD:P_ROUTE_CARD][
            idx * NUMBER_ROAD : (idx + 1) * NUMBER_ROAD
        ][list_road_id_] = 1
    # thẻ route của người chơi
    player_route_card = np.zeros(NUMBER_ROUTE)
    player_route_card[
        np.where(
            env_state[ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN][
                id_action * ATTRIBUTE_PLAYER : (id_action + 1) * ATTRIBUTE_PLAYER
            ][ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :]
            > 0
        )[0]
    ] = 1

    player_state[P_ROUTE_CARD:P_ROUTE_GET] = player_route_card
    # route card get
    route_card_get = env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP].astype(np.int64)
    player_state[P_ROUTE_GET:P_TRAIN_CAR_CARD_BOARD] = route_card_get
    # thẻ train_car trên bàn chơi
    for type_car in env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET]:
        if type_car != -1:
            player_state[P_TRAIN_CAR_CARD_BOARD:P_CARD_BULD_TUNNEL][int(type_car)] += 1
    player_state[P_CARD_BULD_TUNNEL:P_CARD_TEST_TUNNEL] = env_state[
        ENV_CARD_BULD_TUNNEL:ENV_CARD_TEST_TUNNEL
    ]
    player_state[P_ID_ACTION] = 0  # dấu hiệu đánh giá mình hành động
    player_state[P_PHASE + int(env_state[ENV_PHASE]) - 1] = 1  # edit 13h 4/1/2023
    player_state[P_CHECK_ROUTE_CARD] = int(env_state[ENV_ROUTE_CARD_BOARD] > -1)
    player_state[P_CHEKC_END] = int(env_state[ENV_CHECK_END] / 2)
    if player_state[P_CHEKC_END] == 1:
        # điểm sẽ trừ của tất cả người chơi
        all_player_neg_score = env_state[
            ENV_IN4_PLAYER + ATT_NEG_SCORE : ENV_TRAIN_CAR_OPEN : ATTRIBUTE_PLAYER
        ]
        player_state[P_NEG_SCORE:P_TRAIN_CAR_CARD] = np.concatenate(
            (all_player_neg_score[id_action:], all_player_neg_score[:id_action])
        )
        all_check_most_route = env_state[ENV_CHECK_MOST_ROUTE:ENV_CHECK_LONGEST_ROAD]
        all_check_longest_route = env_state[
            ENV_CHECK_LONGEST_ROAD : ENV_CHECK_LONGEST_ROAD + NUMBER_PLAYER
        ]
        player_state[P_CHECK_MOST_ROUTE:P_CHECK_LONGEST_ROAD] = np.concatenate(
            (all_check_most_route[id_action:], all_check_most_route[:id_action])
        )
        player_state[
            P_CHECK_LONGEST_ROAD : P_CHECK_LONGEST_ROAD + NUMBER_PLAYER
        ] = np.concatenate(
            (all_check_longest_route[id_action:], all_check_longest_route[:id_action])
        )

    if env_state[ENV_ID_PLAYER_END] != -1:
        player_state[
            P_ID_PLAYER_END
            + int((env_state[ENV_ID_PLAYER_END] - id_action) % NUMBER_PLAYER)
        ] = 1
    player_state[P_NUMBER_TRAIN_CAR_GET] = env_state[ENV_NUMBER_TRAIN_CAR_GET]
    player_state[P_NUMBER_DROP_ROUTE_CARD] = env_state[ENV_NUMBER_DROP_ROUTE_CARD]
    player_state[P_CARD_TEST_TUNNEL:P_TYPE_TRAIN_CAR_BUILD_ROAD] = env_state[
        ENV_CARD_TEST_TUNNEL : ENV_CARD_TEST_TUNNEL + NUMBER_TYPE_TRAIN_CAR_CARD
    ]
    player_state[
        P_TYPE_TRAIN_CAR_BUILD_ROAD : P_TYPE_TRAIN_CAR_BUILD_ROAD
        + NUMBER_TYPE_TRAIN_CAR_CARD
    ] = env_state[
        ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
        + NUMBER_TYPE_TRAIN_CAR_CARD
    ]
    if (
        env_state[ENV_TRAIN_CAR_CARD] != -1
        and np.min(env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET]) > -1
    ):
        player_state[P_ACTION_GET_TRAIN_CAR_DOWN] = 1
    return player_state


#########################################################


@njit()
def getReward(player_state):
    if player_state[P_CHEKC_END] == 1:
        list_score = (
            player_state[P_SCORE:P_NEG_SCORE]
            - player_state[P_NEG_SCORE:P_TRAIN_CAR_CARD]
        )
        check_highest_score = np.zeros(NUMBER_PLAYER)
        if np.argmax(list_score) == 0:
            list_highest_score = np.where(list_score == np.max(list_score))[0]
            if len(list_highest_score) == 1:
                return 1
            else:
                check_highest_score[list_highest_score] = 1
                check_most_route = (
                    player_state[P_CHECK_MOST_ROUTE:P_CHECK_LONGEST_ROAD]
                    * check_highest_score
                )
                if np.argmax(check_most_route) == 0:
                    if np.sum(check_most_route) == 1:
                        return 1
                    else:
                        check_longest_road = player_state[
                            P_CHECK_LONGEST_ROAD : P_CHECK_LONGEST_ROAD + NUMBER_PLAYER
                        ]

                        if np.sum(check_most_route) == 0:
                            check_longest_road = (
                                check_longest_road * check_highest_score
                            )
                        else:
                            check_longest_road = check_longest_road * check_most_route

                        if check_longest_road[0] == 1:
                            return 1
                        else:
                            if np.sum(check_longest_road) > 0:
                                return 0
                            else:
                                return 1
                else:
                    return 0
        else:
            return 0
    else:
        return -1


#########################################################


@njit()
def check_winner(env_state):
    # B1: tìm người chơi con đường dài nhất
    list_longest_path = np.zeros(NUMBER_PLAYER)
    all_road = env_state[ENV_ROAD_BOARD:ENV_ROUTE_CARD_BOARD]
    number_route_comp = np.zeros(NUMBER_PLAYER)
    for player in range(NUMBER_PLAYER):
        p_road = np.where(all_road == player)[0]
        all_player_route_card = np.where(
            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * player : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (player + 1)
            ][ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :]
            == 1
        )[0]
        # tính toán điểm của thẻ route
        for route_card in all_player_route_card:
            point_source = LIST_ALL_ROUTE_POINT[route_card][0]
            point_dest = LIST_ALL_ROUTE_POINT[route_card][1]
            check_done = check_done_route_card(p_road, point_source, point_dest)
            if check_done == 1:
                number_route_comp[player] += 1
                env_state[
                    ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * player + ATT_SCORE
                ] += LIST_ALL_SCORE_ROUTE[route_card]
            else:
                env_state[
                    ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * player + ATT_NEG_SCORE
                ] += LIST_ALL_SCORE_ROUTE[route_card]

        list_longest_path[player] = calculator_longest_road(p_road)
    # xác định đường dài nhất
    max_road = np.max(list_longest_path)
    player_longest_path = np.where(list_longest_path == max_road)[0]
    # gán đánh dấu người chơi có con đường dài nhất
    env_state[ENV_CHECK_LONGEST_ROAD : ENV_CHECK_LONGEST_ROAD + NUMBER_PLAYER][
        player_longest_path
    ] = 1
    # cộng 10đ cho những người có đường dài nhất
    for player in player_longest_path:
        env_state[ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * player + ATT_SCORE] += 10
    # cập nhật số đường người chơi hoàn thành
    env_state[ENV_CHECK_MOST_ROUTE:ENV_CHECK_LONGEST_ROAD] = number_route_comp
    # xét người chiến thắng
    all_player_score = env_state[
        ENV_IN4_PLAYER : ENV_TRAIN_CAR_OPEN - 1 : ATTRIBUTE_PLAYER
    ]
    all_player_neg_score = env_state[
        ENV_IN4_PLAYER + ATT_NEG_SCORE : ENV_TRAIN_CAR_OPEN - 1 : ATTRIBUTE_PLAYER
    ]
    all_player_score = all_player_score - all_player_neg_score
    score_max = np.max(all_player_score)
    winner_1 = np.where(all_player_score == score_max)[0]
    if len(winner_1) == 1:
        return winner_1, env_state
    else:
        number_route_winner_comp = np.full(NUMBER_PLAYER, -1)
        number_route_winner_comp[winner_1] = number_route_comp[winner_1]
        route_max = np.max(number_route_winner_comp)
        winner_2 = np.where(number_route_winner_comp == route_max)[0]
        #  env_state[ENV_CHECK_MOST_ROUTE : ENV_CHECK_LONGEST_ROAD][winner_2] = 1
        if len(winner_2) == 1:
            #  player_win = winner_2[0]
            return winner_2, env_state
        else:
            winner_3 = np.array([-1])
            for player in winner_2:
                if player in player_longest_path:
                    winner_3 = np.append(winner_3, player)
            if len(winner_3) == 1:
                return winner_2, env_state
            else:
                winner_3 = winner_3[1:]
                #  env_state[ENV_CHECK_LONGEST_ROAD : ENV_CHECK_LONGEST_ROAD + NUMBER_PLAYER][winner_3] = 1
                return winner_3, env_state


#########################################################


@njit()
def system_check_end(env_state):
    if (
        env_state[ENV_CHECK_END] == 2
        and env_state[ENV_ID_PLAYER_END] == env_state[ENV_ID_ACTION]
    ):
        return False
    else:
        return True


#########################################################


@njit()
def getValidActions(player_state_origin):
    player_state = player_state_origin.copy()
    phase_game = np.where(player_state[P_PHASE : P_PHASE + NUMBER_PHASE] == 1)[0][0] + 1
    list_action_return = np.zeros(NUMBER_ACTIONS, dtype=np.float64)
    if phase_game == 1:
        # kiểm tra các hành động nhặt train_car làm được
        train_car_board = player_state[P_TRAIN_CAR_CARD_BOARD:P_CARD_BULD_TUNNEL]
        if player_state[P_ACTION_GET_TRAIN_CAR_DOWN] == 1:
            list_action_return[156] = 1  # action nhặt thẻ úp
        if player_state[P_NUMBER_TRAIN_CAR_GET] == 0:
            # kiểm tra nhặt được thẻ ko
            if player_state[P_CHECK_ROUTE_CARD] != 0:
                list_action_return[168] = 1
            # kiểm tra các đường xây được
            road_can_build = check_road_can_build(player_state)
            list_action_return[road_can_build] = 1
            train_car_can_get = np.where(train_car_board > 0)[0] + 147
            list_action_return[train_car_can_get] = 1
        else:  #  player_state[P_NUMBER_TRAIN_CAR_GET] != 0:
            # nếu đã nhặt thẻ train_car thì k được nhặt thẻ locomotive
            train_car_board[0] = 0
            train_car_can_get = np.where(train_car_board > 0)[0] + 147
            list_action_return[train_car_can_get] = 1
    elif phase_game == 2:
        # chọn tài nguyên để xây đường
        """
        các kiểu dùng tài nguyên để xây con đường muốn xây được lưu ở player_state, từ đó xác định các cách xây có thể
        """
        type_train_car_build_road = np.where(
            player_state[
                P_TYPE_TRAIN_CAR_BUILD_ROAD : P_TYPE_TRAIN_CAR_BUILD_ROAD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ]
            > 0
        )[0]
        list_action_return[type_train_car_build_road + 157] = 1

    elif phase_game == 3:
        route_card_can_drop = np.where(
            player_state[P_ROUTE_GET:P_TRAIN_CAR_CARD_BOARD] == 1
        )[0]
        if player_state[P_NUMBER_DROP_ROUTE_CARD] < 2:
            list_action_return[route_card_can_drop + NUMBER_ROAD] = 1
        list_action_return[169] = 1  # dừng drop thẻ route

    elif phase_game == 4:
        card_build_tunnel = np.where(
            player_state[P_CARD_BULD_TUNNEL:P_CARD_TEST_TUNNEL] > 0
        )[0]
        card_test_tunnel = player_state[P_CARD_TEST_TUNNEL:P_TYPE_TRAIN_CAR_BUILD_ROAD]
        player_train_car = player_state[P_TRAIN_CAR_CARD:P_PLAYER_ROAD]
        check = 1
        test_train_car = 0
        for type_car in card_build_tunnel:
            if card_test_tunnel[type_car] > 0:
                if type_car == 0:
                    test_train_car -= card_test_tunnel[type_car]
                else:
                    test_train_car -= max(
                        card_test_tunnel[type_car] - player_train_car[type_car], 0
                    )
            else:
                continue
        if player_train_car[0] + test_train_car < 0:
            check = 0
        list_action_return[166] = check
        list_action_return[167] = 1
    if np.sum(list_action_return) == 0:
        list_action_return[-1] = 1

    return list_action_return


#########################################################


@njit()
def stepEnv(env_state, action):
    phase_game = env_state[ENV_PHASE]
    id_action = int(env_state[ENV_ID_ACTION])

    if phase_game == 1:
        if action < NUMBER_ROAD:  # nếu action là xây đường, xử lí xây đường
            road = action
            player_train_car = env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][
                ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ]
            temp_train_car = player_train_car.copy()
            train_car_else = np.where(temp_train_car[1:] > 0)[0] + 1
            temp_train_car[train_car_else] += temp_train_car[0]
            # nếu đường ko cần đầu máy
            if LIST_ALL_TYPE_ROAD[road] != 2:
                road_color = LIST_ALL_COLOR_ROAD[road]
                # nếu là đường màu xám
                if road_color == -1:
                    type_build = np.where(temp_train_car >= LIST_ALL_LENGTH_ROAD[road])[
                        0
                    ]
                    env_state[
                        ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
                        + NUMBER_TYPE_TRAIN_CAR_CARD
                    ][type_build] = 1
                else:
                    if player_train_car[0] >= LIST_ALL_LENGTH_ROAD[road]:
                        env_state[
                            ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
                            + NUMBER_TYPE_TRAIN_CAR_CARD
                        ][0] = 1
                    env_state[
                        ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
                        + NUMBER_TYPE_TRAIN_CAR_CARD
                    ][road_color] = 1

            # nếu là ferry cần locomotive
            else:
                type_build = np.where(temp_train_car >= LIST_ALL_LENGTH_ROAD[road])[0]
                env_state[
                    ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][type_build] = 1
            env_state[ENV_PHASE] = 2
            env_state[ENV_ROAD_BUILT] = action
        # nếu nhặt thẻ locomotive thì update thông tin bàn chơi rồi chuyển người
        elif action == 147:
            # cập nhật các thẻ mở

            new_train_car = env_state[ENV_TRAIN_CAR_CARD]
            index_drop = np.where(
                env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET] == 0
            )[0][0]
            env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET][index_drop] = new_train_car
            env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER] = np.concatenate(
                (env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER][1:], np.array([-1]))
            )

            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][action - 147 + ATT_NUMBER_TYPE_TRAIN_CARD] += 1
            env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
            if env_state[ENV_CHECK_END] >= 1:
                if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                    env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023
            env_state = process_train_car_board(env_state)
            # xáo thẻ drop nếu chồng bài úp hết bài
            if env_state[ENV_TRAIN_CAR_CARD] == -1 and (
                np.sum(env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL])
                or env_state[ENV_TRAIN_CAR_CARD] != -1
            ):
                env_state = shuffle_drop_card(env_state)
            # Lật thêm thẻ ra khi thẻ mở đang không đủ 5 thẻ
            if (
                np.min(
                    env_state[
                        ENV_TRAIN_CAR_OPEN : ENV_TRAIN_CAR_OPEN
                        + NUMBER_TRAIN_CAR_CARD_OPEN
                    ]
                )
                == -1
                and env_state[ENV_TRAIN_CAR_CARD] != -1
            ):
                env_state = shuffle_drop_card(env_state)
                env_state = process_train_car_board(env_state)

        # nếu nhặt thẻ thường
        elif action < 156:
            type_car_get = action - 147
            new_train_car = env_state[ENV_TRAIN_CAR_CARD]
            index_drop = np.where(
                env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET] == type_car_get
            )[0][0]
            env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET][index_drop] = new_train_car
            env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER] = np.concatenate(
                (env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER][1:], np.array([-1]))
            )
            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][type_car_get + ATT_NUMBER_TYPE_TRAIN_CARD] += 1
            env_state = process_train_car_board(env_state)

            # xáo thẻ drop nếu chồng bài úp hết bài
            if env_state[ENV_TRAIN_CAR_CARD] == -1 and (
                np.sum(env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL])
                or env_state[ENV_TRAIN_CAR_CARD] != -1
            ):
                env_state = shuffle_drop_card(env_state)
            # Lật thêm thẻ ra khi thẻ mở đang không đủ 5 thẻ
            if (
                np.min(
                    env_state[
                        ENV_TRAIN_CAR_OPEN : ENV_TRAIN_CAR_OPEN
                        + NUMBER_TRAIN_CAR_CARD_OPEN
                    ]
                )
                == -1
                and env_state[ENV_TRAIN_CAR_CARD] != -1
            ):
                env_state = shuffle_drop_card(env_state)
                env_state = process_train_car_board(env_state)
            env_state[ENV_NUMBER_TRAIN_CAR_GET] += 1
            # nếu đã nhặt đủ 2 thẻ
            if env_state[ENV_NUMBER_TRAIN_CAR_GET] == 2:
                # reset số thẻ đã nhặt, cập nhật người chơi mơi
                env_state[ENV_NUMBER_TRAIN_CAR_GET] = 0
                env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
                if env_state[ENV_CHECK_END] >= 1:
                    if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                        env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023
        # nếu nhặt thẻ từ chồng train_Car úp
        elif action == 156:
            new_train_car = int(env_state[ENV_TRAIN_CAR_CARD])
            #  print('màu của thẻ úp được nhặt là ', LIST_COLOR[int(new_train_car)])
            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][new_train_car + ATT_NUMBER_TYPE_TRAIN_CARD] += 1
            env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER] = np.concatenate(
                (env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER][1:], np.array([-1]))
            )
            env_state[ENV_NUMBER_TRAIN_CAR_GET] += 1
            # xáo thẻ drop nếu chồng bài úp hết bài
            if env_state[ENV_TRAIN_CAR_CARD] == -1 and np.sum(
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL]
            ):
                env_state = shuffle_drop_card(env_state)
            # nếu đã nhặt đủ 2 thẻ
            if env_state[ENV_NUMBER_TRAIN_CAR_GET] == 2:
                # reset số thẻ đã nhặt, cập nhật người chơi mơi
                env_state[ENV_NUMBER_TRAIN_CAR_GET] = 0
                env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
                if env_state[ENV_CHECK_END] >= 1:
                    if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                        env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023
        # nếu nhặt thẻ routecard thì cập nhật rồi sang phase 3
        elif action == 168:
            route_card_get = env_state[
                ENV_ROUTE_CARD_BOARD : ENV_ROUTE_CARD_BOARD + NUMBER_ROUTE_GET
            ]
            route_card_get = route_card_get[route_card_get > -1].astype(np.int64)
            env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP][route_card_get] = 1
            env_state[ENV_ROUTE_CARD_BOARD:ENV_TRAIN_CAR_CARD] = np.concatenate(
                (
                    env_state[ENV_ROUTE_CARD_BOARD:ENV_TRAIN_CAR_CARD][
                        NUMBER_ROUTE_GET:
                    ],
                    np.array([-1] * NUMBER_ROUTE_GET),
                )
            )
            env_state[ENV_PHASE] = 3
            if len(route_card_get) == 3:
                env_state[ENV_NUMBER_DROP_ROUTE_CARD] = 0
            else:
                env_state[ENV_NUMBER_DROP_ROUTE_CARD] = 2
        # nếu ko action được gì khác thì skip
        else:
            # action == 170
            # skip qua người chơi
            env_state[ENV_NUMBER_TRAIN_CAR_GET] = 0
            env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
            if env_state[ENV_CHECK_END] >= 1:
                if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                    env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023

    elif phase_game == 2:
        type_train_car_use = action - 157
        road = int(env_state[ENV_ROAD_BUILT])
        type_road = LIST_ALL_TYPE_ROAD[road]
        player_train_car = env_state[
            ENV_IN4_PLAYER
            + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
            + ATTRIBUTE_PLAYER * (id_action + 1)
        ][
            ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
            + NUMBER_TYPE_TRAIN_CAR_CARD
        ]
        # nếu xây road thường
        if type_road == 0:
            if player_train_car[type_train_car_use] >= LIST_ALL_LENGTH_ROAD[road]:
                # bỏ thẻ vào chồng bài drop
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL][
                    type_train_car_use
                ] += LIST_ALL_LENGTH_ROAD[road]
                # đủ train_Car nên chỉ trừ train_Car
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ] -= LIST_ALL_LENGTH_ROAD[
                    road
                ]
            else:
                # nếu k đủ traincar thì tối thiểu việc dùng locomotive
                locomotive_use = (
                    LIST_ALL_LENGTH_ROAD[road] - player_train_car[type_train_car_use]
                )
                # bỏ thẻ vào chồng bài drop
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL][
                    type_train_car_use
                ] += env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ]
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL][
                    type_train_car_use - type_train_car_use
                ] += locomotive_use
                # trừ tài nguyên người chơi
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use - type_train_car_use
                ] -= locomotive_use
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ] = 0

            # cộng điểm cho người chơi, trừ số tàu của người chơi
            env_state[
                ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_SCORE
            ] += LIST_SCORE_BUILD_ROAD[
                LIST_ALL_LENGTH_ROAD[int(env_state[ENV_ROAD_BUILT])]
            ]
            env_state[
                ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_NUMBER_TRAIN
            ] -= LIST_ALL_LENGTH_ROAD[int(env_state[ENV_ROAD_BUILT])]
            if (
                env_state[
                    ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_NUMBER_TRAIN
                ]
                <= 2
                and env_state[ENV_CHECK_END] == 0
            ):
                env_state[ENV_CHECK_END] = 1
                env_state[ENV_ID_PLAYER_END] = (id_action + 1) % NUMBER_PLAYER
            # update đường của người chơi
            env_state[road] = id_action
            env_state[ENV_ROAD_BUILT] = -1
            env_state[
                ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ] = 0
            env_state[ENV_PHASE] = 1
            env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
            if env_state[ENV_CHECK_END] >= 1:
                if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                    env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023
            if (
                np.min(env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET]) == -1
                or env_state[ENV_TRAIN_CAR_CARD] == -1
            ) and np.sum(env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL]) > 0:
                env_state = shuffle_drop_card(env_state)
            # Lật thêm thẻ ra khi thẻ mở đang không đủ 5 thẻ
            if (
                np.min(
                    env_state[
                        ENV_TRAIN_CAR_OPEN : ENV_TRAIN_CAR_OPEN
                        + NUMBER_TRAIN_CAR_CARD_OPEN
                    ]
                )
                == -1
                and env_state[ENV_TRAIN_CAR_CARD] != -1
            ):
                env_state = shuffle_drop_card(env_state)
                env_state = process_train_car_board(env_state)
        # nếu xây ferry
        elif type_road == 2:
            temp_train_car = player_train_car.copy()
            temp_train_car[1:] += temp_train_car[0]
            locomotive_need_use = LIST_ROAD_LOCOMOTIVES[road]
            # player_train_car
            if player_train_car[type_train_car_use] >= LIST_ALL_LENGTH_ROAD[road]:
                # nếu đủ train_Car thì chỉ trừ train_Car
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use - type_train_car_use
                ] -= locomotive_need_use
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ] -= (
                    LIST_ALL_LENGTH_ROAD[road] - locomotive_need_use
                )
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL][
                    type_train_car_use - type_train_car_use
                ] += locomotive_need_use
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL][
                    type_train_car_use
                ] += (LIST_ALL_LENGTH_ROAD[road] - locomotive_need_use)

            else:
                locomotive_need_use += (
                    LIST_ALL_LENGTH_ROAD[road]
                    - locomotive_need_use
                    - player_train_car[type_train_car_use]
                )
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use - type_train_car_use
                ] -= locomotive_need_use
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ] -= (
                    LIST_ALL_LENGTH_ROAD[road] - locomotive_need_use
                )
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL][
                    type_train_car_use - type_train_car_use
                ] += locomotive_need_use
                env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL][
                    type_train_car_use
                ] += (LIST_ALL_LENGTH_ROAD[road] - locomotive_need_use)
            env_state[
                ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_SCORE
            ] += LIST_SCORE_BUILD_ROAD[
                LIST_ALL_LENGTH_ROAD[int(env_state[ENV_ROAD_BUILT])]
            ]
            env_state[
                ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_NUMBER_TRAIN
            ] -= LIST_ALL_LENGTH_ROAD[int(env_state[ENV_ROAD_BUILT])]
            if (
                env_state[
                    ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_NUMBER_TRAIN
                ]
                <= 2
            ):
                env_state[ENV_CHECK_END] = 1
                env_state[ENV_ID_PLAYER_END] = (id_action + 1) % NUMBER_PLAYER
            # update đường của người chơi
            env_state[road] = id_action
            env_state[ENV_ROAD_BUILT] = -1
            env_state[
                ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ] = 0
            env_state[ENV_PHASE] = 1
            env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
            if env_state[ENV_CHECK_END] >= 1:
                if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                    env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023
            if (
                np.min(env_state[ENV_TRAIN_CAR_OPEN:ENV_ROUTE_CARD_GET]) == -1
                or env_state[ENV_TRAIN_CAR_CARD] == -1
            ) and np.sum(env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL]) > 0:
                env_state = shuffle_drop_card(env_state)
            # Lật thêm thẻ ra khi thẻ mở đang không đủ 5 thẻ
            if (
                np.min(
                    env_state[
                        ENV_TRAIN_CAR_OPEN : ENV_TRAIN_CAR_OPEN
                        + NUMBER_TRAIN_CAR_CARD_OPEN
                    ]
                )
                == -1
                and env_state[ENV_TRAIN_CAR_CARD] != -1
            ):
                env_state = shuffle_drop_card(env_state)
                env_state = process_train_car_board(env_state)
        # nếu xây tunnel
        else:
            if player_train_car[type_train_car_use] >= LIST_ALL_LENGTH_ROAD[road]:
                # bỏ thẻ vào chồng bài build_tunnel
                env_state[ENV_CARD_BULD_TUNNEL:ENV_CARD_TEST_TUNNEL][
                    type_train_car_use
                ] += LIST_ALL_LENGTH_ROAD[road]
                # nếu đủ train_Car thì chỉ trừ train_Car
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ] -= LIST_ALL_LENGTH_ROAD[
                    road
                ]
            else:
                # nếu k đủ traincar thì tối thiểu việc dùng locomotive
                locomotive_use = (
                    LIST_ALL_LENGTH_ROAD[road] - player_train_car[type_train_car_use]
                )
                # bỏ thẻ vào chồng bài build_tunnel
                env_state[ENV_CARD_BULD_TUNNEL:ENV_CARD_TEST_TUNNEL][
                    type_train_car_use
                ] += env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ]
                env_state[ENV_CARD_BULD_TUNNEL:ENV_CARD_TEST_TUNNEL][
                    type_train_car_use - type_train_car_use
                ] += locomotive_use
                # trừ tài nguyên người chơi
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use - type_train_car_use
                ] -= locomotive_use
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                    + NUMBER_TYPE_TRAIN_CAR_CARD
                ][
                    type_train_car_use
                ] = 0

            # đi lật bài để test tunnel
            if env_state[ENV_TRAIN_CAR_CARD + 3] == -1:
                env_state = shuffle_drop_card(env_state)
            car_test_tunnel = env_state[
                ENV_TRAIN_CAR_CARD : ENV_TRAIN_CAR_CARD + NUMBER_CARD_TEST_TUNNEL
            ].astype(np.int64)
            #  print('check', car_test_tunnel)
            for car in car_test_tunnel:
                if car != -1:
                    env_state[ENV_CARD_TEST_TUNNEL + car] += 1
            # điều chỉnh chồng bài úp
            env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER] = np.concatenate(
                (
                    env_state[ENV_TRAIN_CAR_CARD:ENV_IN4_PLAYER][
                        NUMBER_CARD_TEST_TUNNEL:
                    ],
                    np.array([-1] * NUMBER_CARD_TEST_TUNNEL),
                )
            )
            env_state[
                ENV_TYPE_TRAIN_CAR_BUILD_ROAD : ENV_TYPE_TRAIN_CAR_BUILD_ROAD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ] = 0
            env_state[ENV_PHASE] = 4

    elif phase_game == 3:
        # nếu dừng bỏ thẻ:
        if action == 169:
            route_card_save = np.where(
                env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP] == 1
            )[0]
            player_route_card = env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :]
            player_route_card[route_card_save] = 1
            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][
                ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :
            ] = player_route_card
            env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP] = 0
            env_state[ENV_NUMBER_DROP_ROUTE_CARD] = 0
            env_state[ENV_PHASE] = 1
            env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
            if env_state[ENV_CHECK_END] >= 1:
                if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                    env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023

            if env_state[ENV_TURN] <= NUMBER_PLAYER:
                env_state[ENV_TURN] += 1
            if env_state[ENV_TURN] <= NUMBER_PLAYER:
                id_action = int(env_state[ENV_ID_ACTION])
                env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP] = env_state[
                    ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN
                ][id_action * ATTRIBUTE_PLAYER : (id_action + 1) * ATTRIBUTE_PLAYER][
                    3 + NUMBER_TYPE_TRAIN_CAR_CARD :
                ]
                env_state[ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN][
                    id_action * ATTRIBUTE_PLAYER : (id_action + 1) * ATTRIBUTE_PLAYER
                ][2 + NUMBER_TYPE_TRAIN_CAR_CARD :] = 0
                env_state[ENV_PHASE] = 3
        else:
            drop_card = action - 101
            # lưu thẻ mình đã bỏ nếu ko phải lượt bỏ thẻ ban đầu
            if env_state[ENV_TURN] > NUMBER_PLAYER:
                route_card_board = env_state[ENV_ROUTE_CARD_BOARD:ENV_TRAIN_CAR_CARD]
                route_insert = np.where(route_card_board == -1)[0][0]
                route_card_board[route_insert] = drop_card
                env_state[ENV_ROUTE_CARD_BOARD:ENV_TRAIN_CAR_CARD] = route_card_board
            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :][drop_card] = -1
            env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP][drop_card] = 0
            env_state[ENV_NUMBER_DROP_ROUTE_CARD] += 1
            # nếu ko được bỏ thẻ nữa
            if env_state[ENV_NUMBER_DROP_ROUTE_CARD] == 2:
                route_card_save = np.where(
                    env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP] == 1
                )[0]
                player_route_card = env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :]
                player_route_card[route_card_save] = 1
                env_state[
                    ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                    + ATTRIBUTE_PLAYER * (id_action + 1)
                ][
                    ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :
                ] = player_route_card
                env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP] = 0
                env_state[ENV_NUMBER_DROP_ROUTE_CARD] = 0
                env_state[ENV_PHASE] = 1
                env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER

                if env_state[ENV_CHECK_END] >= 1:
                    if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                        env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023
                #  env_state[ENV_TURN] += 1
                if env_state[ENV_TURN] <= NUMBER_PLAYER:
                    env_state[ENV_TURN] += 1
                if env_state[ENV_TURN] <= NUMBER_PLAYER:
                    env_state[ENV_PHASE] = 3
                    id_action = int(env_state[ENV_ID_ACTION])
                    env_state[ENV_ROUTE_CARD_GET:ENV_TRAIN_CAR_DROP] = env_state[
                        ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN
                    ][
                        id_action
                        * ATTRIBUTE_PLAYER : (id_action + 1)
                        * ATTRIBUTE_PLAYER
                    ][
                        3 + NUMBER_TYPE_TRAIN_CAR_CARD :
                    ]
                    env_state[ENV_IN4_PLAYER:ENV_TRAIN_CAR_OPEN][
                        id_action
                        * ATTRIBUTE_PLAYER : (id_action + 1)
                        * ATTRIBUTE_PLAYER
                    ][ATT_NUMBER_TYPE_TRAIN_CARD + NUMBER_TYPE_TRAIN_CAR_CARD :] = 0

    elif phase_game == 4:
        train_car_return = env_state[ENV_CARD_BULD_TUNNEL:ENV_CARD_TEST_TUNNEL]
        test_train_car = env_state[ENV_CARD_TEST_TUNNEL:ENV_PHASE]
        if action == 167:  # không xây hầm
            # trả tài nguyên cho người chơi đã bỏ ra để định xây hầm, nạp các thẻ test xây hầm vào chồng bài úp
            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][
                ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ] += train_car_return
            env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL] += test_train_car

        elif action == 166:  # có xây hầm
            # kiểm tra xem trùng loại thẻ nào, trừ loại thẻ đấy rồi cho hết lá bài vào drop
            temp = train_car_return * test_train_car
            subtract_car = (temp >= 1) * test_train_car
            # trừ tài nguyên nếu trùng vs tài nguyên test tunnel
            train_car_player = env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][
                ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ]
            train_car_player -= subtract_car
            for type_train_car in range(1, NUMBER_TYPE_TRAIN_CAR_CARD):
                if train_car_player[type_train_car] < 0:
                    train_car_player[0] += train_car_player[type_train_car]
                    train_car_player[type_train_car] = 0
            env_state[
                ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * id_action : ENV_IN4_PLAYER
                + ATTRIBUTE_PLAYER * (id_action + 1)
            ][
                ATT_NUMBER_TYPE_TRAIN_CARD : ATT_NUMBER_TYPE_TRAIN_CARD
                + NUMBER_TYPE_TRAIN_CAR_CARD
            ] = train_car_player
            env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL] += (
                test_train_car + train_car_return + subtract_car
            )
            # cộng điểm cho người chơi
            env_state[
                ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_SCORE
            ] += LIST_SCORE_BUILD_ROAD[
                LIST_ALL_LENGTH_ROAD[int(env_state[ENV_ROAD_BUILT])]
            ]
            env_state[
                ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_NUMBER_TRAIN
            ] -= LIST_ALL_LENGTH_ROAD[int(env_state[ENV_ROAD_BUILT])]
            if (
                env_state[
                    ENV_IN4_PLAYER + ATTRIBUTE_PLAYER * id_action + ATT_NUMBER_TRAIN
                ]
                <= 2
                and env_state[ENV_CHECK_END] == 0
            ):
                env_state[ENV_CHECK_END] = 1
                env_state[ENV_ID_PLAYER_END] = (id_action + 1) % NUMBER_PLAYER
            # ghi nhận đường cho người chơi:
            env_state[int(env_state[ENV_ROAD_BUILT])] = id_action
        # update phase, chuyển người chơi
        env_state[ENV_PHASE] = 1
        env_state[ENV_CARD_BULD_TUNNEL:ENV_PHASE] = 0
        env_state[ENV_ID_ACTION] = (id_action + 1) % NUMBER_PLAYER
        if env_state[ENV_CHECK_END] >= 1:
            if env_state[ENV_ID_ACTION] == env_state[ENV_ID_PLAYER_END]:
                env_state[ENV_CHECK_END] += 0.5  # edit by Hieu 05012023
        if (
            env_state[ENV_TRAIN_CAR_CARD] == -1
            and np.sum(env_state[ENV_TRAIN_CAR_DROP:ENV_CARD_BULD_TUNNEL]) > 0
        ):
            env_state = shuffle_drop_card(env_state)
        # Lật thêm thẻ ra khi thẻ mở đang không đủ 5 thẻ
        if (
            np.min(
                env_state[
                    ENV_TRAIN_CAR_OPEN : ENV_TRAIN_CAR_OPEN + NUMBER_TRAIN_CAR_CARD_OPEN
                ]
            )
            == -1
            and env_state[ENV_TRAIN_CAR_CARD] != -1
        ):
            env_state = shuffle_drop_card(env_state)
            env_state = process_train_car_board(env_state)
    return env_state


#  #########################################################


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    idx = np.random.randint(0, arr_action.shape[0])
    return arr_action[idx], perData


#########################################################


@njit()
def one_game_numba(p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    env_state = initEnv()
    count_turn = 0

    while system_check_end(env_state) and count_turn < 3000:
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            if getValidActions(player_state)[action] != 1:
                raise Exception("bot dua ra action khong hop le")

        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")

        env_state = stepEnv(env_state, action)
        count_turn += 1

    winner, env_state = check_winner(env_state)
    for p_idx in range(5):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER

    if np.where(list_other == -1)[0][0] in winner:
        winner = True
    else:
        winner = False
    return winner, per_player


@njit()
def n_games_numba(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner
    return win, per_player


def one_game_normal(p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4):
    env_state = initEnv()
    count_turn = 0

    while system_check_end(env_state) and count_turn < 3000:
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            if getValidActions(player_state)[action] != 1:
                raise Exception("bot dua ra action khong hop le")

        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")

        env_state = stepEnv(env_state, action)
        count_turn += 1

    winner, env_state = check_winner(env_state)

    for p_idx in range(5):
        env_state[ENV_PHASE] = 1
        idx = int(env_state[ENV_ID_ACTION])
        player_state = getAgentState(env_state)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        elif list_other[idx] == 4:
            action, per4 = p4(player_state, per4)
        else:
            raise Exception("Sai list_other.")
        env_state[ENV_ID_ACTION] = (env_state[ENV_ID_ACTION] + 1) % NUMBER_PLAYER
    if np.where(list_other == -1)[0][0] in winner:
        winner = True
    else:
        winner = False
    return winner, per_player


def n_games_normal(
    p0, num_game, per_player, list_other, per1, per2, per3, per4, p1, p2, p3, p4
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, per4, p1, p2, p3, p4
        )
        win += winner
    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


@njit()
def check_run_under_njit(agent, perData):
    return True


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
        )



import numpy as np
from PIL import Image, ImageDraw, ImageFont

from src.Base.Sheriff import env as _env
from src.Base.Sheriff.docs import index
from env import SHORT_PATH

IMG_PATH = SHORT_PATH + "src/Base/Sheriff/images/"
import os


class Sprites:
    def __init__(self) -> None:
        self.background = Image.open(IMG_PATH + "background.png").resize((2100, 900))
        self._background_ = self.background.copy()
        self.sheriff = (
            Image.open(IMG_PATH + "sheriff.png").resize((100, 130)).convert("RGBA")
        )
        self.cards = {}
        card_names = os.listdir(IMG_PATH)
        for name in card_names:
            if name != "background.png":
                self.cards[name.split(".png")[0]] = (
                    Image.open(IMG_PATH + name).resize((100, 130)).convert("RGBA")
                )


sprites = Sprites()


class Params:
    def __init__(self) -> None:
        self.card_order = [
            "apples",
            "cheese",
            "bread",
            "chicken",
            "pepper",
            "mead",
            "silk",
            "crossbow",
            "2-apples",
            "2-cheese",
            "2-bread",
            "2-chicken",
            "3-apples",
            "3-cheese",
            "3-bread",
        ]
        self.font28 = ImageFont.FreeTypeFont("src/ImageFonts/arial.ttf", 28)


params = Params()


def draw_outlined_text(draw, text, font, pos, color, opx):
    o_color = (255 - color[0], 255 - color[1], 255 - color[2])
    x = pos[0]
    y = pos[1]
    draw.text((x + opx, y + opx), text, o_color, font)
    draw.text((x - opx, y + opx), text, o_color, font)
    draw.text((x + opx, y - opx), text, o_color, font)
    draw.text((x - opx, y - opx), text, o_color, font)
    draw.text(pos, text, color, font)


def get_state_image(state=None):
    if state is None:
        return sprites.background

    bg = sprites.background.copy()

    # Draw
    draw = ImageDraw.ImageDraw(bg)

    # Số coin đang có, số coin nợ, ai là sheriff
    id_sheriff = np.where(state[[2, 102, 131, 160]] == 1)[0]
    if len(id_sheriff) == 1:
        id_sheriff = int(id_sheriff[0])
    else:
        id_sheriff = -1

    if id_sheriff != -1:
        bg.paste(sprites.sheriff, (415 + 500 * id_sheriff, 210), sprites.sheriff)

    s_0 = 380
    for p in range(4):
        s_ = 85 + 500 * p
        if p == 0:
            text = "coin: " + str(int(state[0]))
            text_1 = "debt: " + str(int(state[1]))
            text_2 = str(int(state[7]))
            text_3 = str(int(state[8]))
            text_4 = str(int(state[9]))
        else:
            text = "coin: " + str(int(state[100 + 29 * (p - 1)]))
            text_1 = "debt: " + str(int(state[101 + 29 * (p - 1)]))
            text_2 = str(int(state[107 + 29 * (p - 1)]))
            text_3 = str(int(state[108 + 29 * (p - 1)]))
            text_4 = str(int(state[109 + 29 * (p - 1)]))

        draw_outlined_text(draw, text, params.font28, (s_, 20), (0, 0, 0), 1)
        bbox = draw.textbbox((0, 0), text_1, params.font28)
        draw_outlined_text(
            draw, text_1, params.font28, (s_ + 430 - bbox[2], 20), (0, 0, 0), 1
        )
        bbox = draw.textbbox((0, 0), text_2, params.font28)
        draw_outlined_text(
            draw, text_2, params.font28, (s_ + 317 - bbox[2], 210), (0, 0, 0), 1
        )
        bbox = draw.textbbox((0, 0), text_3, params.font28)
        draw_outlined_text(
            draw, text_3, params.font28, (s_ + 317 - bbox[2], 260), (0, 0, 0), 1
        )
        bbox = draw.textbbox((0, 0), text_4, params.font28)
        draw_outlined_text(
            draw, text_4, params.font28, (s_ + 317 - bbox[2], 310), (0, 0, 0), 1
        )

        if p == 0:
            temp = state[3:7].astype(int)
        else:
            temp = state[103 + 29 * (p - 1) : 107 + 29 * (p - 1)].astype(int)

        for i in range(4):
            text = str(temp[i] * int(state[474 + p]))
            draw_outlined_text(
                draw, text, params.font28, (s_ + 3 + 110 * i, 70), (255, 255, 255), 1
            )

        if p != 0:
            temp = state[110 + 29 * (p - 1) : 125 + 29 * (p - 1)].astype(int)
            for i in range(15):
                text = str(temp[i])
                bbox = draw.textbbox((0, 0), text, params.font28)
                draw_outlined_text(
                    draw,
                    text,
                    params.font28,
                    (s_0 + 110 * i + 55 - bbox[2], 600 + 35 * (p - 1)),
                    (255, 255, 255),
                    1,
                )

            temp = state[410 + 15 * p : 425 + 15 * p].astype(int)
            for i in range(15):
                text = str(temp[i])
                draw_outlined_text(
                    draw,
                    text,
                    params.font28,
                    (s_0 + 110 * i - 15, 600 + 35 * (p - 1)),
                    (255, 255, 255),
                    1,
                )

            temp = state[125 + 29 * (p - 1) : 129 + 29 * (p - 1)].astype(int)
            for i in range(4):
                text = str(temp[i])
                draw_outlined_text(
                    draw,
                    text,
                    params.font28,
                    (s_ + 3 + 110 * i, 170),
                    (255, 255, 255),
                    1,
                )

    for k in range(3):
        inf_1 = state[10 + 30 * k : 25 + 30 * k].astype(int)
        inf_0 = state[25 + 30 * k : 40 + 30 * k].astype(int)
        for i in range(15):
            draw_outlined_text(
                draw,
                str(inf_0[i]),
                params.font28,
                (s_0 + 110 * i - 15, 775 + 35 * k),
                (255, 255, 255),
                1,
            )
            text = str(inf_1[i])
            bbox = draw.textbbox((0, 0), text, params.font28)
            draw_outlined_text(
                draw,
                text,
                params.font28,
                (s_0 + 110 * i + 55 - bbox[2], 775 + 35 * k),
                (255, 255, 255),
                1,
            )

    temp = state[367:382].astype(int)
    for i in range(15):
        text = str(temp[i])
        draw_outlined_text(
            draw, text, params.font28, (s_0 + 110 * i - 30, 702), (255, 255, 255), 1
        )

    temp = state[382:397].astype(int)
    for i in range(15):
        text = str(temp[i])
        bbox = draw.textbbox((0, 0), text, params.font28)
        draw_outlined_text(
            draw,
            text,
            params.font28,
            (s_0 + 110 * i + 50 - bbox[2] // 2 - 30, 710),
            (255, 255, 255),
            1,
        )

    temp = state[397:412].astype(int)
    for i in range(15):
        text = str(temp[i])
        bbox = draw.textbbox((0, 0), text, params.font28)
        draw_outlined_text(
            draw,
            text,
            params.font28,
            (s_0 + 110 * i + 100 - bbox[2] - 30, 718),
            (255, 255, 255),
            1,
        )

    left = []
    right = []
    for i in range(6):
        try:
            a = np.where(state[187 + 15 * i : 202 + 15 * i] == 1)[0][0]
            left.append(params.card_order[a])
        except:
            pass

        try:
            a = np.where(state[277 + 15 * i : 292 + 15 * i] == 1)[0][0]
            right.append(params.card_order[a])
        except:
            pass

    w_ = len(left) * 110 - 10
    s_ = 600 - w_ // 2
    for i in range(len(left)):
        bg.paste(sprites.cards[left[i]], (s_ + 110 * i, 400))

    w_ = len(right) * 110 - 10
    s_ = 1500 - w_ // 2
    for i in range(len(right)):
        bg.paste(sprites.cards[right[i]], (s_ + 110 * i, 400))

    phase = np.where(state[414:425] == 1)[0][0]
    text = str(phase)
    bbox = draw.textbbox((0, 0), text, params.font28)
    draw_outlined_text(
        draw, text, params.font28, (1050 - bbox[2] // 2, 450), (0, 0, 0), 1
    )
    return bg


action_annotations = [
    "bỏ thẻ apple",
    "bỏ thẻ cheese",
    "bỏ thẻ bread",
    "bỏ thẻ chicken",
    "bỏ thẻ peper",
    "bỏ thẻ mead",
    "bỏ thẻ silk",
    "bỏ thẻ crossbow",
    "bỏ thẻ green_apple",
    "bỏ thẻ gouda_cheese",
    "bỏ thẻ rye_bread",
    "bỏ thẻ royal_rooster",
    "bỏ thẻ golden_apple",
    "bỏ thẻ bleu_cheese",
    "bỏ thẻ pump_bread",
    "Không bỏ thẻ nữa",
    "Lấy thẻ chồng bài rút",
    "Lấy thẻ chồng bài lật trái",
    "Lấy thẻ chồng bài lật phải",
    "Trả thẻ vào chồng bài lật trái",
    "Trả thẻ vào chồng bài lật phải",
    "bỏ thẻ apple vào túi",
    "bỏ thẻ cheese vào túi",
    "bỏ thẻ bread vào túi",
    "bỏ thẻ chicken vào túi",
    "bỏ thẻ peper vào túi",
    "bỏ thẻ mead vào túi",
    "bỏ thẻ silk vào túi",
    "bỏ thẻ crossbow vào túi",
    "bỏ thẻ green_apple vào túi",
    "bỏ thẻ gouda_cheese vào túi",
    "bỏ thẻ rye_bread vào túi",
    "bỏ thẻ royal_rooster vào túi",
    "bỏ thẻ golden_apple vào túi",
    "bỏ thẻ bleu_cheese vào túi",
    "bỏ thẻ pump_bread vào túi",
    "Không bỏ thẻ vào túi nữa",
    "Khai báo hàng là apple",
    "Khai báo hàng là cheese",
    "Khai báo hàng là bread",
    "Khai báo hàng là chicken",
    "Kiểm tra người đầu tiên cạnh mình",
    "Kiểm tra người thứ 2 cạnh mình",
    "Kiểm tra người thứ 3 cạnh mình",
    "Không hối lộ coin nữa",
    "Hối lộ thêm 1 coin",
    "hối lộ thẻ apple done",
    "hối lộ thẻ cheese done",
    "hối lộ thẻ bread done",
    "hối lộ thẻ chicken done",
    "hối lộ thẻ peper done",
    "hối lộ thẻ mead done",
    "hối lộ thẻ silk done",
    "hối lộ thẻ crossbow done",
    "hối lộ thẻ green_apple done",
    "hối lộ thẻ gouda_cheese done",
    "hối lộ thẻ rye_bread done",
    "hối lộ thẻ royal_rooster done",
    "hối lộ thẻ golden_apple done",
    "hối lộ thẻ bleu_cheese done",
    "hối lộ thẻ pump_bread done",
    "Không hối lộ thẻ done nữa",
    "hối lộ thẻ apple trong túi",
    "hối lộ thẻ cheese trong túi",
    "hối lộ thẻ bread trong túi",
    "hối lộ thẻ chicken trong túi",
    "hối lộ thẻ peper trong túi",
    "hối lộ thẻ mead trong túi",
    "hối lộ thẻ silk trong túi",
    "hối lộ thẻ crossbow trong túi",
    "hối lộ thẻ green_apple trong túi",
    "hối lộ thẻ gouda_cheese trong túi",
    "hối lộ thẻ rye_bread trong túi",
    "hối lộ thẻ royal_rooster trong túi",
    "hối lộ thẻ golden_apple trong túi",
    "hối lộ thẻ bleu_cheese trong túi",
    "hối lộ thẻ pump_bread trong túi",
    "Không hối lộ thẻ trong túi nữa",
    "Kiểm tra hàng",
    "Cho qua",
    "Bỏ thẻ tịch thu vào bên trái",
    "Bỏ thẻ tịch thu vào bên phải",
]


def get_description(action):
    if action < 0 or action >= _env.getActionSize():
        return ""

    return action_annotations[action]


class Env_components:
    def __init__(self, env, count_turn, list_other) -> None:
        self.env = env
        self.count_turn = count_turn
        self.list_other = list_other


def get_env_components():
    env = _env.initEnv()
    count_turn = 0
    list_other = np.array([-1, 1, 2, 3])
    np.random.shuffle(list_other)
    return Env_components(env, count_turn, list_other)


def get_main_player_state(
    env_components: Env_components, list_agent, list_data, action=None
):
    if not action is None:
        env_components.env = _env.stepEnv(env_components.env, action)
        env_components.count_turn += 1

    while (
        _env.system_check_end(env_components.env) and env_components.count_turn < 1000
    ):
        p_idx = int(env_components.env[index.ENV_ID_ACTION])
        if env_components.list_other[p_idx] == -1:
            break

        state = _env.getAgentState(env_components.env)
        agent = list_agent[env_components.list_other[p_idx] - 1]
        data = list_data[env_components.list_other[p_idx] - 1]
        action, data = agent(state, data)
        env_components.env = _env.stepEnv(env_components.env, action)
        env_components.count_turn += 1

    if _env.system_check_end(env_components.env) and env_components.count_turn < 1000:
        state = _env.getAgentState(env_components.env)
        win = -1
    else:
        env = env_components.env.copy()
        env[index.ENV_PHASE] = 1
        env[index.ENV_CHECK_END] = 1
        my_idx = np.where(env_components.list_other == -1)[0][0]
        env[index.ENV_ID_ACTION] = my_idx
        state = _env.getAgentState(env)
        if _env.check_winner(env_components.env) == my_idx:
            win = 1
        else:
            win = 0

        for p_idx in range(4):
            if p_idx != my_idx:
                env[index.ENV_ID_ACTION] = p_idx
                _state = _env.getAgentState(env)
                agent = list_agent[env_components.list_other[p_idx] - 1]
                data = list_data[env_components.list_other[p_idx] - 1]
                action, data = agent(_state, data)

    return win, state, env_components



import numpy as np
from numba import njit

# Gofish


@njit
def getActionSize():
    return 17


@njit
def getAgentSize():
    return 4


@njit
def getStateSize():
    return 107


@njit
def initEnv():
    all = np.arange(52) % 13
    np.random.shuffle(all)

    env = np.zeros(170)  ### --- update
    env[:52] = all

    for i in range(20):
        idx = i % 4
        laBai = int(env[i])
        env[53 + idx * 15 + laBai] += 1

    env[52] = 32
    for i in range(4):
        arr = env[53 + 15 * i : 66 + 15 * i]
        env[67 + i * 15] = np.where(arr == 4)[0].size
        env[66 + i * 15] = 5 - env[67 + i * 15] * 4

    env[116] = -1
    return env


@njit
def getAgentState(env):
    state = np.zeros(107)

    idx = int(env[113] % 4)  # player chính
    state[:15] = env[53 + idx * 15 : 68 + idx * 15]
    for k in range(3):
        pidx = int((idx + k + 1) % 4)
        arr_ = env[53 + pidx * 15 : 66 + pidx * 15]
        arr = state[15 + k * 15 : 30 + k * 15]
        arr[arr_ == 4] += 1
        arr[13:15] = env[66 + pidx * 15 : 68 + pidx * 15]

        state[67 + k * 13 : 80 + k * 13] = env[
            117 + pidx * 13 : 130 + pidx * 13
        ]  ### --- update

    state[60] = env[52]  ### số lá còn lại

    state[61 + int(env[114])] = 1  # phase
    if env[115]:
        state[64 + int(env[115]) - 1] = 1  # người bị yêu cầu

    state[-1] = env[-1]
    return state


@njit
def getValidActions(state):
    validActions = np.zeros(17)
    arrPhase = np.where(state[61:64] == 1)[0]
    if arrPhase.size:
        phase = arrPhase[0]
        if phase == 0:
            arr = state[28:60:15]
            idxs = np.where(arr > 0)[0]
            validActions[idxs + 1] += 1
        elif phase == 1:
            arr = state[0:13]
            idx = np.where((arr > 0) & (arr < 4))[0]
            validActions[idx + 4] = 1
        elif phase == 2:
            validActions[0] = 1

    if state[-1] == 1:
        validActions = np.ones(17)

    return validActions


@njit
def stepEnv(action, env):
    player_0 = env[113] % 4
    arr_0 = env[53 + 15 * player_0 : 68 + 15 * player_0]

    if action == 0:  # Bốc 1 là bài từ bộ bài
        laBai = int(env[52 - int(env[52])])
        env[52] -= 1
        arr_0[laBai] += 1
        arr_0[13] += 1

        # tinh lai diem
        if arr_0[laBai] == 4:
            arr_0[14] += 1  # điểm tăng 1
            arr_0[13] -= 4

        # neu het bai tren tay thi phai lay du 5 la tu bo bai
        if arr_0[13] == 0:
            while env[52] > 0 and arr_0[13] < 5:
                l_ = int(env[52 - int(env[52])])
                env[52] -= 1
                arr_0[l_] += 1
                arr_0[13] += 1
            # Tính điểm
            laBaicuaToi = arr_0[:13]
            arr_0[13] = np.sum(laBaicuaToi[laBaicuaToi < 4])
            arr_0[14] = np.where(laBaicuaToi == 4)[0].size

        if laBai == env[116] and arr_0[13] != 0:
            env[114:117] = np.array([0, 0, -1])
        else:
            if np.sum(env[66:113:15]) - arr_0[13] != 0:
                env[114:117] = np.array([3, 0, -1])
            elif env[52] != 0:
                env[114:117] = np.array([2, 0, -1])

    elif action < 4:
        env[114] = 1
        env[115] = action
    elif action < 17:
        laBai = int(action - 4)
        player_1 = int(player_0 + env[115]) % 4
        arr_1 = env[53 + 15 * player_1 : 68 + 15 * player_1]

        env[117 + int(player_0) * 13 + laBai] = 1  ###--- update

        if (
            arr_1[laBai] > 0
        ):  ### không cần nhỏ hơn 4 vì nếu người này đủ thì chắc chắn người kia không yêu cầu được
            arr_0[laBai] += arr_1[laBai]  # đưa thẻ
            arr_0[13] += arr_1[laBai]
            arr_1[13] -= arr_1[laBai]
            arr_1[laBai] = 0

            # tinh lai diem
            if arr_0[laBai] == 4:
                arr_0[14] += 1  # điểm tăng 1
                arr_0[13] -= 4

            # neu het bai tren tay thi phai lay du 5 la tu bo bai
            if arr_0[13] == 0:
                while env[52] > 0 and arr_0[13] < 5:
                    l_ = int(env[52 - int(env[52])])
                    env[52] -= 1
                    arr_0[l_] += 1
                    arr_0[13] += 1
                # tinh diem
                laBaicuaToi = arr_0[:13]
                arr_0[13] = np.sum(laBaicuaToi[laBaicuaToi < 4])
                arr_0[14] = np.where(laBaicuaToi == 4)[0].size

            if arr_0[13] != 0:
                if np.sum(env[66:113:15]) - arr_0[13] != 0:
                    env[114:117] = np.array([0, 0, -1])
                elif env[52]:
                    env[114:117] = np.array([2, 0, -1])
            else:
                env[114:117] = np.array([3, 0, -1])

        else:
            if env[52]:
                env[114:117] = np.array([2, 0, laBai])
            else:
                env[114:117] = np.array([3, 0, -1])

    # phase = 3---------------
    if np.sum(env[67:113:15]) != 13:
        while env[114] == 3:
            env[113] += 1  # chuyển sang người chơi khác
            new_pl = int(env[113] % 4)

            # Nếu người tiếp theo hết bài và trên bàn còn bài ---> thì bốc đủ cho họ
            new_arr = env[53 + new_pl * 15 : 68 + new_pl * 15]
            if env[52] and new_arr[13] == 0:
                while env[52] > 0 and new_arr[13] < 5:
                    l_ = int(env[52 - int(env[52])])
                    env[52] -= 1
                    new_arr[l_] += 1
                    new_arr[13] += 1
                # tinh diem
                laBaicuaHo = new_arr[:13]
                new_arr[13] = np.sum(laBaicuaHo[laBaicuaHo < 4])
                new_arr[14] = np.where(laBaicuaHo == 4)[0].size

            if new_arr[13] != 0:  # người chơi này còn bài
                env[114:117] = np.array([0, 0, -1])

            env[117 + new_pl * 13 : 130 + new_pl * 13] = np.zeros(13)  ###--- update
    else:
        env[-1] = 1


@njit
def checkEnded(env):
    if env[-1] == 1:
        scoreArr = env[67:113:15]
        max = np.max(scoreArr)
        if np.where(scoreArr == max)[0].size == 1:
            winner = np.argmax(scoreArr)
        else:
            laMax = 0
            for i in range(4):
                if env[67 + i * 15] == max:
                    arr_player = env[53 + i * 15 : 66 + i * 15]
                    l = np.where(arr_player == 4)[0][-1]
                    if l > laMax:
                        laMax = l
                        winner = i
        return winner
    else:
        return -1


@njit
def getReward(state):
    if state[-1] == 0:
        return -1
    else:
        arrMax = state[29:60:15]
        max_ = np.max(arrMax)
        if state[14] > max_:
            return 1
        elif state[14] == max_:
            laMax = 0
            for i in range(3):
                if arrMax[i] == max_:
                    arr = state[15 + i * 15 : 28 + i * 15]
                    l = np.where(arr == 1)[0][-1]
                    if laMax < l:
                        laMax = l
            if laMax < np.where(state[:13] == 4)[0][-1]:
                return 1
    return 0


# def visualizeEnv(env):
#   print('ENV-------')
#   print('Những lá còn lại:',env[52], env[52- int(env[52]): 52] + 1)
#   for i in range(4):
#     print('--> p:',i, np.where((env[53 + 15*i: 66+ 15*i] > 0) & (env[53 + 15*i: 66+ 15*i] < 4) )[0] + 1, env[66+ 15*i: 68+ 15*i])
#     print('arr', env[53 + 15*i: 66+ 15*i] )

#   print(env[113: 117])
#   print(env[117: 169].reshape(4,13))
#   print('--- PLAYER:', env[113] %4)

# def visualizeState(state):
#   print('arr: ', state[: 13] , np.where((state[:13] > 0) & (state[:13]< 4) )[0] + 1, state[13: 15])
#   for i in range(3):
#     print('-p:', np.where( state[15 + i*15: 28 + i*15] )[0] + 1, state[28 + i*15: 30 + i*15])
#   print('so la con lai:', state[60])
#   print('phase:', state[61: 64])
#   print('ngBiYeuCau:', state[64: 67])
#   print(state[67: 106].reshape(3,13))
#   print('endgame:', state[106])


@njit
def bot_lv0(state, per):
    validActions = getValidActions(state)
    actions = np.where(validActions == 1)[0]
    action = np.random.choice(actions)
    return action, per


def one_game_normal(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env = initEnv()
    # check = 0
    while env[113] < 400:
        idx = int(env[113]) % 4
        player_state = getAgentState(env)

        # if env[113] == check:
        #   check += 1
        #   print('-------------------------------------')
        # visualizeEnv(env)
        # visualizeState(player_state)

        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            list_action = getValidActions(player_state)
            if list_action[action] != 1:
                raise Exception("Action không hợp lệ")
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        else:
            raise Exception("Sai list_other.")

        stepEnv(action, env)
        winner = checkEnded(env)
        if winner != -1:
            break

    env[-1] = 1
    p0_idx = np.where(list_other == -1)[0][0]
    for p_idx in range(4):
        env[113] = p_idx
        p_state = getAgentState(env)
        if list_other[p_idx] == -1:
            action, per_player = p0(p_state, per_player)
        elif list_other[p_idx] == 1:
            action, per1 = p1(p_state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(p_state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(p_state, per3)
        else:
            raise Exception("Sai list_other.")

    if p0_idx == winner:
        result = 1
    else:
        result = 0
    # print(list_other)
    # print(env[53: 113].reshape(4,15))
    return result, per_player


def n_games_normal(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


@njit
def one_game_numba(p0, list_other, per_player, per1, per2, per3, p1, p2, p3):
    env = initEnv()
    while env[113] < 400:
        idx = int(env[113]) % 4
        player_state = getAgentState(env)
        if list_other[idx] == -1:
            action, per_player = p0(player_state, per_player)
            list_action = getValidActions(player_state)
            if list_action[action] != 1:
                raise Exception("Action không hợp lệ")
        elif list_other[idx] == 1:
            action, per1 = p1(player_state, per1)
        elif list_other[idx] == 2:
            action, per2 = p2(player_state, per2)
        elif list_other[idx] == 3:
            action, per3 = p3(player_state, per3)
        else:
            raise Exception("Sai list_other.")

        stepEnv(action, env)
        winner = checkEnded(env)
        if winner != -1:
            break

    env[-1] = 1
    p0_idx = np.where(list_other == -1)[0][0]
    for p_idx in range(4):
        env[113] = p_idx
        p_state = getAgentState(env)
        if list_other[p_idx] == -1:
            action, per_player = p0(p_state, per_player)
        elif list_other[p_idx] == 1:
            action, per1 = p1(p_state, per1)
        elif list_other[p_idx] == 2:
            action, per2 = p2(p_state, per2)
        elif list_other[p_idx] == 3:
            action, per3 = p3(p_state, per3)
        else:
            raise Exception("Sai list_other.")

    if p0_idx == winner:
        result = 1
    else:
        result = 0
    return result, per_player


@njit
def n_games_numba(p0, num_game, per_player, list_other, per1, per2, per3, p1, p2, p3):
    win = 0
    for _ in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, p1, p2, p3
        )
        win += winner

    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


@njit()
def check_run_under_njit(agent, perData):
    return True


def load_agent(level, *args):
    num_bot = getAgentSize() - 1
    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]

            for i in range(num_bot):
                if level == -1:
                    module_agent = load_module_player(
                        lst_agent_level[i], game_name=env_name
                    )
                    _list_per_level_.append(module_agent.DataAgent())
                else:
                    data_agent_level = np.load(
                        f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                        allow_pickle=True,
                    )
                    module_agent = load_module_player(lst_agent_level[i])
                    _list_per_level_.append(
                        module_agent.convert_to_test(data_agent_level)
                    )
                _list_bot_level_.append(module_agent.Test)

    return _list_bot_level_, _list_per_level_


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0, per_player)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_games_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )
    else:
        return n_games_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
        )



game_name = "Splendor"
time_run_game = 100
N_AGENT = 5
N_GAME = 15
PASS_LEVEL = 2500
COUNT_TRAIN = 10000
COUNT_TEST = 1000

#  path = "C:\AutomaticTrain\State.xlsx"
#  SHOT_PATH = 'A:/AutoTrain/GAME/'
DRIVE_FOLDER = "G:/My Drive/AutomaticColab/"


SHORT_PATH = ""
#  DRIVE_FOLDER = 'H:/Drive của tôi/AutomaticColab/'

import importlib.util
import sys


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module



import sys
import numpy as np
from numba import njit
import env

game_name = sys.argv[1]
env.make(game_name)


from numba.typed import List


def DataAgent():
    perx_ = List()
    stateSize = env.getStateSize()
    actionSize = env.getActionSize()
    perx_.append(np.zeros((1000 * stateSize, actionSize)))
    perx_.append(np.zeros((1000 * stateSize, actionSize)))
    perx_.append(np.zeros((1, 2)))

    temp = np.arange(actionSize, dtype=np.float64)
    perx_.append(np.array([temp]))
    return perx_


@njit
def Train(state, per):
    weight = per[3][0]
    np.random.shuffle(weight)
    actions = env.getValidActions(state)
    output = (weight + 1) * actions
    action = np.argmax(output)

    #  Tìm các max value cho array bias
    if per[2][0][1] >= 9000:
        state_int = state.astype(np.int64)
        state_int[state_int >= 1000] = per[2][0][0] + 1
        where_ = np.where((state_int <= per[2][0][0]) & (state_int >= 0))[0]
        arr_idx = (where_ * 1000 + state_int[where_]).astype(np.int64)
        per[1][arr_idx] += weight
    else:
        max_ = np.max(state)
        if max_ > np.int64(per[2][0][0]):
            per[2][0][0] = max_ + 1

    #  Bắt đầu lưu array bias
    reward = env.getReward(state)
    if per[2][0][1] >= 9000:
        if reward == 1:
            per[0] += per[1]
            per[1][:, :] = 0.0
        elif reward == 0:
            per[1][:, :] = 0.0

    if reward != -1:
        per[2][0][1] += 1

    return action, per


@njit
def Test(state, per):
    state_int = state.astype(np.int64)
    stateSize = env.getStateSize()
    actionSize = env.getActionSize()
    where_ = np.where((state_int < per[stateSize][0]) & (state_int >= 0))[0]
    weight = np.zeros(actionSize)
    for i in where_:
        weight += per[i][state_int[i]]

    actions = env.getValidActions(state)
    output = (weight + 1) * actions

    #  action = np.argmax(output)

    list_action = np.where(actions == 1)[0]
    action = list_action[np.argmax(output[list_action])]
    return action, per


def convert_to_save(per_data):
    if len(per_data) == env.getStateSize() + 1:
        raise Exception("Data này đã được convert rồi.")

    data = List()
    arr = np.zeros(env.getStateSize(), np.int64)
    for i in range(env.getStateSize()):
        for j in range(1000):
            if (per_data[0][1000 * i + j] == 0).all():
                check = True
                for k in range(j, 1000):
                    if (per_data[0][1000 * i + k] != 0).any():
                        check = False
                        break
                if check:
                    arr[i] = j
                    break
        else:
            arr[i] = 1000

    for i in range(env.getStateSize()):
        data.append(per_data[0][1000 * i : 1000 * i + arr[i]])

    data.append(np.array([arr.astype(float)]))
    for i in range(0, len(data) - 1):
        for j in range(data[i].shape[0]):
            if np.max(data[i][j]) != 0.0:
                data[i][j] /= np.max(data[i][j])

    return data


def convert_to_test(per_data):
    return List(per_data)



import sys

import numba
import numpy as np
from numba import jit, njit
from numba.typed import List


@njit
def getActionSize():
    return 93


@njit
def getStateSize():
    return 145


@njit
def getAgentSize():
    return 6


@njit
def initEnv():
    """init env"""
    env = np.zeros(116)
    card = np.arange(65)  # card except Defuse and Explo kitten and Implo Kitten
    np.random.shuffle(card)
    env[0:76] = 7  #  7 is card on draw pile
    for i in range(6):  #  draw 7 card for player: id from 0 to 5
        env[card[i * 7 : (i + 1) * 7]] = i
        env[65 + i] = i
    draw_pile = np.where(env[0:76] == 7.0)[0].astype(np.float64)
    np.random.shuffle(draw_pile)
    discard_pile = np.zeros(19)  # card on discard pile will have id 6

    env[76] = 0  #  nope count
    env[77] = 0  #  track player id main turn
    env[78:83] = [2, 3, 4, 5, 0]  #  track player id Nope turn
    env[83:89] = 1  #  0 if lose else 1
    env[
        89
    ] = 0  # phase [0:main turn, 1:nope turn,2:steal card turn,3:choose/take card turn,4:alter future turn,5:discard turn]
    env[90] = 1  #  number of card player env[77] have to draw
    env[91:94] = [-1, -1, -1]  # three card in see the future
    env[94] = -1  #  player env[77] last action
    env[95] = env[77] + 1
    env[96] = -1  # player chosen in phase 2

    env[97] = 1  #  main direction
    env[98] = 0  #  state of the Imploding Kitten

    env[99] = 0
    env[100:116] = 0
    return env, draw_pile, discard_pile


@njit
def getNumCard(env, idx):
    """Get the number of card with given type"""
    return np.where(env == idx)[0].shape[0]


@njit
def getAllNumCard(env, idx):
    """Get all the number of card"""
    state = np.zeros(17)
    state[0] = getNumCard(env[0:5], idx)
    for i in range(4):
        state[1 + i] = getNumCard(env[5 + i * 4 : 9 + i * 4], idx)
    state[5] = getNumCard(env[21:26], idx)
    for i in range(9):
        state[6 + i] = getNumCard(env[26 + i * 4 : 30 + i * 4], idx)
    state[15] = getNumCard(env[62:65], idx)
    state[16] = getNumCard(env[65:71], idx)
    return state


@njit
def getCardType(id):
    cards = List(
        [
            np.arange(0.0, 5.0),
            np.arange(5.0, 9.0),
            np.arange(9.0, 13.0),
            np.arange(13.0, 17.0),
            np.arange(17.0, 21.0),
            np.arange(21.0, 26.0),
            np.arange(26.0, 30.0),
            np.arange(30.0, 34.0),
            np.arange(34.0, 38.0),
            np.arange(38.0, 42.0),
            np.arange(42.0, 46.0),
            np.arange(46.0, 50.0),
            np.arange(50.0, 54.0),
            np.arange(54.0, 58.0),
            np.arange(58.0, 62.0),
            np.arange(62.0, 65.0),
            np.arange(65.0, 71.0),
            np.arange(71.0, 75.0),
            np.arange(75.0, 76.0),
        ]
    )
    i = 0
    for c in cards:
        if id in c:
            return i
        i += 1


@njit
def visualCard(card):
    arr = []
    lst = [
        "Nope",
        "Attack",
        "Skip",
        "Favor",
        "Shuffle",
        "See the future",
        "TCT",
        "RRC",
        "BC",
        "HPC",
        "CTM",
        "Reverse",
        "Draw Bottom",
        "Feral Cat",
        "Alter the Future",
        "Targeted Attack",
        "Defuse",
        "Exploding Kitten",
        "Imploding Kitten",
    ]
    for i in card:
        if i != -1:
            arr.append(lst[int(getCardType(i))])
    return arr


@njit
def getCardRange(type_card):
    """Get the range of the card given type"""
    cards = List(
        [
            np.arange(0.0, 5.0),
            np.arange(5.0, 9.0),
            np.arange(9.0, 13.0),
            np.arange(13.0, 17.0),
            np.arange(17.0, 21.0),
            np.arange(21.0, 26.0),
            np.arange(26.0, 30.0),
            np.arange(30.0, 34.0),
            np.arange(34.0, 38.0),
            np.arange(38.0, 42.0),
            np.arange(42.0, 46.0),
            np.arange(46.0, 50.0),
            np.arange(50.0, 54.0),
            np.arange(54.0, 58.0),
            np.arange(58.0, 62.0),
            np.arange(62.0, 65.0),
            np.arange(65.0, 71.0),
            np.arange(71.0, 75.0),
            np.arange(75.0, 76.0),
        ]
    )
    return (
        cards[type_card].astype(np.int64)[0],
        cards[type_card].astype(np.int64)[-1] + 1,
    )


@njit
def getAgentState(env, draw_pile, discard_pile):
    state = np.zeros(getStateSize())
    phase = env[89]
    main_id = env[77]
    nope_id = env[95]
    last_action = env[94]
    if phase == 0:
        pIdx = int(main_id)
    elif phase == 1:
        pIdx = int(nope_id)
    elif phase == 2:
        pIdx = int(main_id)
    elif phase == 3:
        if last_action == 3:
            pIdx = int(env[96])
        else:
            pIdx = int(main_id)
    elif phase == 4:
        pIdx = int(main_id)
    elif phase == 5:
        pIdx = int(main_id)
    state[0:17] = getAllNumCard(env, pIdx)
    state[17:35] = discard_pile[:18]  # discard pile
    state[35] = np.where(draw_pile != -1)[0].shape[0]  # number of card in draw pile
    state[36] = np.where(env[83:89] == 1)[0].shape[0]
    state[37] = env[76] % 2  # 1 if action been Nope else 0

    if phase != 5:
        state[95:100][int(phase)] = 1  # phase

    if last_action >= 0:
        state[101:116][int(last_action)] = 1  #  player main turn last action

    for i in range(5):
        state[122 + i] = np.where(env[0:76] == env[78 + i])[0].shape[0]

    state[127] = discard_pile[18]

    if phase == 1:  # Nope phase
        nope_turn = nopeTurn(nope_id, reverse=(env[97] == 0))
        for i in range(5):
            state[116 + i] = env[83:89][int(nope_turn[i])]
        state[121] = env[83:89][int(nope_id)]  # lose or not

    elif phase == 3 and last_action == 3:  #
        nope_turn = nopeTurn(env[96], reverse=(env[97] == 0))
        for i in range(5):
            state[116 + i] = env[83:89][int(nope_turn[i])]
        state[121] = env[83:89][int(env[96])]  # lose or not

    else:
        if phase == 5:  # discard phase
            state[128:144] = env[100:116]
            state[144] = env[99]

        if pIdx == int(main_id):
            for i in range(3):
                if env[91 + i] != -1:
                    card = np.zeros(19)
                    card[int(getCardType(env[91 + i]))] = 1
                    state[
                        38 + 19 * i : 57 + 19 * i
                    ] = card  #  three card if use see the future
        state[100] = np.maximum(env[90], 0)  #  number of card player have to draw

        for i in range(5):
            state[116 + i] = env[83:89][int(env[78 + i])]
        state[121] = env[83:89][int(env[77])]  # lose or not

    return state


@njit
def getValidActions(state):
    list_action = np.zeros(getActionSize())
    if state[95] == 1:  # main turn
        list_action[1:6] = (state[1:6] > 0).astype(np.float64)
        if np.sum(state[122:127]) == 0:
            list_action[3] = 0
        list_action[6] = 1
        if np.max(state[0:16]) >= 2 and np.sum(state[122:127]) > 0:  # two of a kind
            list_action[7] = 1
        elif np.max(state[6:11]) + state[13] >= 2 and np.sum(state[122:127]) > 0:
            list_action[7] = 1

        if np.max(state[0:16]) >= 3 and np.sum(state[122:127]) > 0:  # three of a kind
            list_action[8] = 1
        elif np.max(state[6:11]) + state[13] >= 3 and np.sum(state[122:127]) > 0:
            list_action[8] = 1

        type_card = (state[0:16] > 0).astype(np.float64)
        type_card[13] = 0
        if (np.sum(type_card) + state[13] >= 5) and np.sum(
            state[17:34]
        ) > 0:  # five of a kind
            list_action[9] = True
        list_action[11:13] = (state[11:13] > 0).astype(np.float64)  # 4 new action
        list_action[13:15] = (state[14:16] > 0).astype(np.float64)

        if np.sum(list_action) == 0:
            list_action[10] = 1
    elif state[96] == 1:  # Nope turn
        if state[0] > 0:
            list_action[0] = 1  # Nope
        list_action[10] = 1  # skip Nope

    elif state[97] == 1:  # Choose player
        for i in range(5):
            if state[116 + i] == 1 and state[122 + i] > 0:
                list_action[15 + i] = 1
        if np.sum(list_action[15:20]) == 0:
            list_action[6] = 1

    elif state[98] == 1:  # choose/take card turn
        main_action = np.where(state[101:116] == 1)[0][0]
        if main_action == 3:
            list_action[20:37][np.where(state[0:17] > 0)] = 1
        elif main_action == 8:
            list_action[37:54] = 1
        elif main_action == 9:
            list_action[54:70] = ((state[17:33] - state[128:144]) > 0) * 1.0
            list_action[70] = (state[33] > 0) * 1.0  # defuse
    elif state[99] == 1:  # Alter future phase
        ##list_change = np.array([[0,1,2], [0,2,1], [1,0,2], [1,2,0], [2,0,1], [2,1,0]])
        if state[35] >= 3:
            list_action[71:77] = 1
        elif state[35] == 2:
            list_action[71] = 1
            list_action[73] = 1
        else:
            list_action[71] = 1
    elif np.sum(state[95:100]) == 0:  #  discard phase
        last_action = np.argmax(state[101:116])
        if last_action == 7:
            if np.max(state[128:144]) == 0:  # discard no card yet
                list_action[77:93][6:11] = (
                    state[6:11] >= max(2 - state[13], 1)
                ).astype(np.float64)
                list_action[77:93][13] = (state[13] > 0) * 1.0

                list_action[77:93][0:6] = (state[0:6] >= 2).astype(np.float64)
                list_action[77:93][11:16] = (state[11:16] >= 2).astype(np.float64)

            elif np.max(state[128:144]) > 0:  # one card have been discard
                type_card = np.argmax(state[128:144])  # card that have been discarded
                if type_card == 13 or (
                    type_card >= 6 and type_card < 11
                ):  # feral cat chosen
                    list_action[77:93][type_card] = (state[type_card] > 0) * 1.0
                    list_action[77:93][13] = (state[13] > 0) * 1.0
                else:
                    list_action[77:93][type_card] = 1

        elif last_action == 8:
            if np.max(state[128:144]) == 0:  # discard no card yet
                list_action[77:93][6:11] = (
                    state[6:11] >= max(3 - state[13], 1)
                ).astype(np.float64)
                list_action[77:93][13] = (state[13] > 0) * 1.0

                list_action[77:93][0:6] = (state[0:6] >= 3).astype(np.float64)
                list_action[77:93][11:16] = (state[11:16] >= 3).astype(np.float64)

            elif np.max(state[128:144]) > 0:  # one card have been discard
                type_card = np.argmax(state[128:144])  # card that have been discarded
                if type_card == 13 or (
                    type_card >= 6 and type_card < 11
                ):  # feral cat chosen
                    list_action[77:93][type_card] = (state[type_card] > 0) * 1.0
                    list_action[77:93][13] = (state[13] > 0) * 1.0
                else:
                    list_action[77:93][type_card] = 1
        elif last_action == 9:
            available_card = (state[0:16] > 0).astype(np.float64)
            list_action[77:93] = ((available_card - state[128:144]) > 0).astype(
                np.float64
            )
            list_action[77:93][13] = (state[13] > 0) * 1.0

    return list_action


@njit
def checkDefuse(env, discard_pile):  #  get the Defuse (if player have else -1)
    card = np.where(env[65:71] == env[77])[0].astype(np.int64)
    if card.shape[0] > 0:
        card_id = card[0]
        env[65:71][card_id] = 6
        discard_pile[16] += 1
        #  print('Player ',env[77],' have Defuse!')
        return True
    return False


@njit
def checkExploding(card):  #  check if that card is expode or not
    explode = np.array([71.0, 72.0, 73.0, 74.0], dtype=np.float64)
    if card in explode:
        #  print('Player draw an Exploding kitten!')
        return True
    return False


@njit
def checkImploding(card):
    # return true if player draw an Imploding Kitten.
    if card == 75:
        #  print('Player draw an Imploding Kitten!')
        return True
    return False


@njit
def nopeTurn(idx, reverse=False):
    if reverse == False:
        return np.arange(idx + 1.0, idx + 6.0) % 6
    else:
        return np.arange(idx + 5.0, idx, -1) % 6


@njit
def changeTurn(env, num_card_draw=1, reverse=False):
    """Change the main turn"""
    if reverse == False:
        env[77] = int(env[77] + 1) % 6
        while env[83:89][int(env[77])] == 0:  # if player id is already lost.
            env[77] = int(env[77] + 1) % 6
        env[78:83] = nopeTurn(env[77], env[97] == 0)
        env[76] = 0  # reset nope count
    else:
        env[77] = int(env[77] - 1) % 6
        while env[83:89][int(env[77])] == 0:  # if player id is already lost.
            env[77] = int(env[77] - 1) % 6
        env[78:83] = nopeTurn(env[77], env[97] == 0)
        env[76] = 0  # reset nope count
    for i in range(5):
        if env[83:89][int(env[78:83][i])] == 1:
            env[95] = env[78:83][i]  # reset nope player id
    #  if env[90]:
    if env[90] >= 2:
        env[90] += num_card_draw  # card next player draw
    else:
        env[90] = num_card_draw
    env[89] = 0  #  change phase to 0
    env[91:94] = -1
    env[94] = -1
    return env


@njit
def drawCard(env, draw_pile, discard_pile, from_bottom=False, change_turn=True):
    """Draw card"""
    #  print('Player ',env[77],' draw ', env[90],'card(s)')
    if from_bottom == True:
        num_cards = 1
    else:
        num_cards = env[90]
    for i in range(int(num_cards)):
        if from_bottom == True:
            index_draw = np.where(draw_pile != -1)[0][-1]
        else:
            index_draw = np.where(draw_pile != -1)[0][0]
        #  print(f'Draw : {visualCard([draw_pile[index_draw]])[0]}')
        if checkExploding(draw_pile[index_draw]):  # draw an exploding kitten
            if checkDefuse(env, discard_pile):  # player have defuse
                idx = np.random.randint(index_draw, draw_pile.shape[0])
                draw_pile_2 = np.zeros_like(draw_pile)
                draw_pile_2[0:index_draw] = draw_pile[0:index_draw]
                draw_pile_2[index_draw:idx] = draw_pile[index_draw + 1 : idx + 1]
                draw_pile_2[idx] = draw_pile[index_draw]
                draw_pile_2[idx + 1 :] = draw_pile[idx + 1 :]
                draw_pile = draw_pile_2
                # insert explode card back to the Draw Pile
            else:  # player lost
                #  print('Player ',env[77],' loss!')
                env[83:89][int(env[77])] = 0
                env[0:76][np.where(env[0:76] == env[77])] = 6
                env[71:75][np.where(env[71:75] != 6)] = 6
                discard_pile[17] += 1
                draw_pile[index_draw] = -1
                change_turn = True
                break
        elif checkImploding(draw_pile[index_draw]):
            if env[98] == 0:  # face down
                env[98] = 1
                idx = np.random.randint(index_draw, draw_pile.shape[0])
                draw_pile_2 = np.zeros_like(draw_pile)
                draw_pile_2[0:index_draw] = draw_pile[0:index_draw]
                draw_pile_2[index_draw:idx] = draw_pile[index_draw + 1 : idx + 1]
                draw_pile_2[idx] = draw_pile[index_draw]
                draw_pile_2[idx + 1 :] = draw_pile[idx + 1 :]
                draw_pile = draw_pile_2
            else:  # face up
                #  print('Player ',env[77],' loss!')
                env[83:89][int(env[77])] = 0
                env[0:76][np.where(env[0:76] == env[77])] = 6
                env[75] = 6
                discard_pile[18] += 1
                draw_pile[index_draw] = -1
                change_turn = True
                break

        else:  # draw other card
            env[0:76][int(draw_pile[index_draw])] = env[77]  # draw
            draw_pile[index_draw] = -1

    if change_turn == True:
        env[90] = 0
        env = changeTurn(env, 1, env[97] == 0)
    return env, draw_pile, discard_pile


@njit
def checkIfNope(env):
    """Return True if the main player's card has been Nope"""
    return env[76] % 2 == 1


@njit
def executeMainAction(env, draw_pile, discard_pile, action):
    """Execute main action if it has not been Nope"""
    #  print('Execute main Action!')
    env[76] = 0
    if action == 1:  # Attack
        #  print(f'Phase {env[89]} Player {env[77]} attack!')
        env[89] = 0
        env = changeTurn(
            env, num_card_draw=2
        )  # change main turn, next player draw 2 card
    elif action == 2:  # Skip
        #  print(f'Phase {env[89]} Player {env[77]} skip!')
        env[90] -= 1
        env[89] = 0
        if env[90] == 0:
            env = changeTurn(env, num_card_draw=1)
    elif action == 3:
        #  print(f'Phase {env[89]} Player {env[77]} use favor!')
        env[89] = 2
    elif action == 4:  # Shuffle
        #  print(f'Phase {env[89]} Player {env[77]} shuffle!')
        np.random.shuffle(draw_pile)
        env[89] = 0
        env[91:94] = -1
    elif action == 5:  # See the future
        #  print(f'Phase {env[89]} Player {env[77]} see the future!')
        if np.where(draw_pile != -1)[0].shape[0] >= 3:
            env[91:94] = draw_pile[np.where(draw_pile != -1)[0][0:3]]
        else:
            env[91:94] = np.concatenate(
                (
                    draw_pile[np.where(draw_pile != -1)[0][0:3]],
                    np.zeros(3 - np.where(draw_pile != -1)[0].shape[0]) - 1,
                )
            )
        env[89] = 0
    elif action == 7:
        #  print(f'Phase {env[89]} Player {env[77]} use two of a kind!')
        env[89] = 2
    elif action == 8:
        #  print(f'Phase {env[89]} Player {env[77]} use three of a kind!')
        env[89] = 2
    elif action == 9:
        #  print(f'Phase {env[89]} Player {env[77]} use five different cards!')
        env[89] = 3
    elif action == 11:  # Reverse
        env[97] = (env[97] + 1) % 2  # change the direction
        env[90] -= 1
        if env[90] == 0:
            changeTurn(env, 1, env[97] == 0)
        env[89] = 0
        #  print(f'Phase {env[89]} Player {env[77]} use Reverse!')
    elif action == 12:  # Draw from bottom
        #  print(f'Phase {env[89]} Player {env[77]} use Draw from Bottom!')
        if env[90] > 1:
            env, draw_pile, discard_pile = drawCard(
                env, draw_pile, discard_pile, from_bottom=True, change_turn=False
            )
            env[90] -= 1
            env[89] = 0
        else:
            env, draw_pile, discard_pile = drawCard(
                env, draw_pile, discard_pile, from_bottom=True, change_turn=True
            )
    elif action == 13:  # Alter the future
        #  print(f'Phase {env[89]} Player {env[77]} use Alter the future!')
        if np.where(draw_pile != -1)[0].shape[0] >= 3:
            env[91:94] = draw_pile[np.where(draw_pile != -1)[0][0:3]]
        else:
            env[91:94] = np.concatenate(
                (
                    draw_pile[np.where(draw_pile != -1)[0][0:3]],
                    np.zeros(3 - np.where(draw_pile != -1)[0].shape[0]) - 1,
                )
            )
        env[89] = 4  # special phase of alter the future
    elif action == 14:  #  Targeted Attack
        #  print(f'Phase {env[89]} Player {env[77]} use Targeted Attack!')
        env[89] = 2  # choose player to attack

    return env, draw_pile, discard_pile


@njit
def discardCardNormalAction(env, last_action, discard_pile):
    if last_action == 0:
        discard_pile[0] += 1
        env[0:5][np.where(env[0:5] == env[77])[0][0]] = 6
    elif last_action == 1:  # Attack
        env[5:9][np.where(env[5:9] == env[77])[0][0]] = 6
        discard_pile[1] += 1
    elif last_action == 2:  # Skip
        env[9:13][np.where(env[9:13] == env[77])[0][0]] = 6
        discard_pile[2] += 1
    elif last_action == 3:
        discard_pile[3] += 1
        env[13:17][np.where(env[13:17] == env[77])[0][0]] = 6
    elif last_action == 4:  # Shuffle
        env[17:21][np.where(env[17:21] == env[77])[0][0]] = 6
        discard_pile[4] += 1
    elif last_action == 5:  # See the future
        env[21:26][np.where(env[21:26] == env[77])[0][0]] = 6
        discard_pile[5] += 1
    elif last_action == 11:  # Reverse
        env[46:50][np.where(env[46:50] == env[77])[0][0]] = 6
        discard_pile[11] += 1
    elif last_action == 12:  # Draw from Bottom
        env[50:54][np.where(env[50:54] == env[77])[0][0]] = 6
        discard_pile[12] += 1
    elif last_action == 13:  # Alter the future
        env[58:62][np.where(env[58:62] == env[77])[0][0]] = 6
        discard_pile[14] += 1
    elif last_action == 14:  # Targeted Attack
        env[62:65][np.where(env[62:65] == env[77])[0][0]] = 6
        discard_pile[15] += 1


@njit
def idPlayerCanUseNope(env, nope_id, reverse=False):
    """return the id of the player that have the nope card, else -1"""
    main_id = env[77]
    nope_turn = nopeTurn(main_id, reverse=(env[97] == 0))

    idx_old = -1
    for i in range(5):
        if nope_turn[i] == nope_id:
            idx_old = i
            break
    else:
        idx_old = -1
    if idx_old + 1 == 5:
        return main_id
    else:
        for i in range(idx_old + 1, 5):
            idx = nope_turn[i]
            if np.where(env[0:5] == idx)[0].shape[0] >= 1 and env[83:89][int(idx)] == 1:
                return idx
        return main_id


@njit
def stepEnv(env, draw_pile, discard_pile, action):
    phase = env[89]
    main_id = env[77]
    nope_id = env[95]
    nope_count = env[76]
    last_action = env[94]
    if phase == 0:  # Phase 0: Main Turn
        env[76] = 0
        if action == 6:  # draw card
            env, draw_pile, discard_pile = drawCard(env, draw_pile, discard_pile)
        else:
            env[94] = action
            if env[94] >= 7 and env[94] <= 9:  # special combo
                env[89] = 5
                if env[94] == 7:
                    env[99] = 2
                elif env[94] == 8:
                    env[99] = 3
                elif env[94] == 9:
                    env[99] = 5
            elif (env[94] >= 0 and env[94] < 7) or (env[94] >= 11 and env[94] < 15):
                discardCardNormalAction(env, env[94], discard_pile)
                env[100:116] = 0
                env[95] = idPlayerCanUseNope(env, main_id)
                if env[95] == main_id:
                    env, draw_pile, discard_pile = executeMainAction(
                        env, draw_pile, discard_pile, env[94]
                    )
                    #  print(f'Phase: {0} Action {action} has been executed!')
                else:
                    env[89] = 1  # nope phase

    elif phase == 1:  # Phase 1: Nope phase
        if action == 0 and env[95] != main_id:  # other player use Nope
            #  print(f'Phase {phase} Player {env[95]} use Nope!')
            env[76] += 1  #  increase Nope Count
            env[0:5][np.where(env[0:5] == env[95])[0][0]] = 6
            discard_pile[0] += 1
            env[95] = idPlayerCanUseNope(env, env[95], env[97] == 0)
            if env[95] == main_id:
                if not checkIfNope(env):  # if not been Nope
                    env, draw_pile, discard_pile = executeMainAction(
                        env, draw_pile, discard_pile, last_action
                    )
                    #  print(f'Phase {phase} Action {last_action} has been executed!')
        elif action == 0 and env[95] == main_id:
            env[76] += 1  #  increase Nope Count
            env[0:5][np.where(env[0:5] == env[77])[0][0]] = 6
            if not checkIfNope(env):  # if not been Nope
                env, draw_pile, discard_pile = executeMainAction(
                    env, draw_pile, discard_pile, last_action
                )
                #  print(f'Phase {0} Action {last_action} has been executed!')
        else:
            if env[95] == main_id:
                if not checkIfNope(env):  # if not been Nope
                    env, draw_pile, discard_pile = executeMainAction(
                        env, draw_pile, discard_pile, last_action
                    )
                    #  print(f'Phase {0} Action {last_action} has been executed!')
                else:  #  if Nope
                    if action == 0:
                        #  print('Main player use Yup!')
                        env[76] = 0  # reset to original
                        env[0:5][np.where(env[0:5] == main_id)[0][0]] = 6
                        discard_pile[0] += 1
                        env[95] = idPlayerCanUseNope(env, env[95], env[97] == 0)
                        if env[95] == main_id:
                            env, draw_pile, discard_pile = executeMainAction(
                                env, draw_pile, discard_pile, last_action
                            )
                            #  print(f'Phase {0} Action {last_action} has been executed!')
                    else:
                        #  print(f'Phase {0} Action {last_action} has been Nope!')
                        env[94] = -1  #  action has been Nope
                        env[89] = 0  #  back to phase 0
                        env[95] = idPlayerCanUseNope(env, main_id, env[97] == 0)
            #  env[95] = idPlayerCanUseNope(env,env[95])
            else:
                env[95] = idPlayerCanUseNope(env, env[95], env[97] == 0)
                if env[95] == main_id:
                    if not checkIfNope(env):  # if not been Nope
                        env, draw_pile, discard_pile = executeMainAction(
                            env, draw_pile, discard_pile, env[94]
                        )
                        #  print(f'Phase {0} Action {env[94]} has been executed!')

    elif (
        phase == 2
    ):  #  phase 2: choose player to steal card. Only main_id can enter this phase
        if action == 6:
            env, draw_pile, discard_pile = drawCard(env, draw_pile, discard_pile)
        else:
            env[96] = env[78:83][int(action - 15)]
            last_action = env[94]
            #  print(f'Phase {phase} Player {env[77]} choose player {env[96]} to steal!')
            if last_action == 7:
                card_on_player_chosen = np.where(env[0:76] == env[96])[0]
                card = np.random.choice(card_on_player_chosen)
                env[0:76][card] = env[77]
                # used card go to Discard Pile
                env[94] = -1  # reset last action
                env[89] = 0
            elif last_action == 14:
                env[77] = env[96]
                env[90] = 2
                env[89] = 0
                env[94] = -1  # reset last action
                env[91:94] = -1
            else:
                env[89] = 3
        env[76] = 0

    elif (
        phase == 3
    ):  # phase 3: choose card to give/take. Only main_id can enter this phase
        last_action = env[94]
        if last_action == 3:
            type_card = action - 20
            low, high = getCardRange(type_card)
            all_card_to_take = np.where(env[low:high] == env[96])[0]
            env[low:high][int(all_card_to_take[0])] = env[77]

        elif last_action == 8:
            # take card
            type_card = action - 37
            low, high = getCardRange(type_card)
            all_card_to_take = np.where(env[low:high] == env[96])[0]
            if all_card_to_take.shape[0] > 0:
                env[low:high][int(all_card_to_take[0])] = env[77]
            # used card go to Discard Pile
        elif last_action == 9:
            type_card = action - 54
            low, high = getCardRange(type_card)
            if np.where(env[low:high] == 6)[0].shape[0] > 0:
                env[low:high][np.where(env[low:high] == 6)[0][0]] = env[77]
                discard_pile[int(type_card)] -= 1
        env[94] = -1  # reset last action
        env[89] = 0
        env[76] = 0

    elif phase == 4:  # Special phase: Alter the future
        act = int(action - 71)
        list_change = np.array(
            [[0, 1, 2], [0, 2, 1], [1, 0, 2], [1, 2, 0], [2, 0, 1], [2, 1, 0]]
        )
        env[91:94] = env[91:94][list_change[act]]
        index_future = np.where(draw_pile != -1)[0]
        if index_future.shape[0] >= 3:
            if act > 0:
                draw_pile[index_future[0:3]] = env[91:94]
        elif index_future.shape[0] == 2:
            if act == 2:
                draw_pile[index_future] = draw_pile[index_future][np.array([1, 0])]
        env[94] = -1
        env[89] = 0
        env[76] = 0

    elif phase == 5:
        type_card = action - 77
        low, high = getCardRange(type_card)  # range
        env[low:high][np.where(env[low:high] == main_id)[0][0]] = 6
        discard_pile[int(type_card)] += 1
        env[100:116][int(type_card)] += 1
        env[99] -= 1

        if env[99] == 0:
            env[100:116] = 0
            env[95] = idPlayerCanUseNope(env, main_id)
            if env[95] == main_id:
                env, draw_pile, discard_pile = executeMainAction(
                    env, draw_pile, discard_pile, env[94]
                )
                #  print(f'Phase: {0} Action {last_action} has been executed!')
                #  print('ok')
            else:
                env[89] = 1  # nope phase

    return env, draw_pile, discard_pile


@njit
def checkEnded(env):
    if np.sum(env[83:89]) == 1:
        return np.where(env[83:89] == 1)[0][0]
    else:
        return -1


@njit
def getReward(state):
    if state[121] == 0:
        return 0
    elif np.sum(state[116:121]) == 0:
        return 1
    else:
        return -1


@njit
def random_player(state, per):
    list_action = np.where(getValidActions(state) == 1)[0]
    action = np.random.choice(list_action)
    # if list_action.shape[0]==0:
    #     print(list(state))
    #  print(list_action)
    return action, per


@njit()
def bot_lv0(state, perData):
    validActions = getValidActions(state)
    arr_action = np.where(validActions == 1)[0]
    # if arr_action.shape[0]==0:
    #     print(list(state))
    idx = np.random.randint(0, arr_action.shape[0])

    return arr_action[idx], perData


@njit
def one_game_numba(
    p0, pIdOrder, per_player, per1, per2, per3, per4, per5, p1, p2, p3, p4, p5
):
    env, draw_pile, discard_pile = initEnv()

    winner = -1
    turn = 0
    while True:
        turn += 1
        phase = env[89]
        main_id = env[77]
        nope_id = env[95]
        last_action = env[94]
        if phase == 0:
            pIdx = int(main_id)
        elif phase == 1:
            pIdx = int(nope_id)
        elif phase == 2:
            pIdx = int(main_id)
        elif phase == 3:
            if last_action == 3:
                pIdx = int(env[96])
            else:
                pIdx = int(main_id)
        elif phase == 4:
            pIdx = int(main_id)
        elif phase == 5:
            pIdx = int(main_id)
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(
                getAgentState(env, draw_pile, discard_pile), per_player
            )
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
        elif pIdOrder[pIdx] == 4:
            action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)
        elif pIdOrder[pIdx] == 5:
            action, per5 = p5(getAgentState(env, draw_pile, discard_pile), per5)
        env, draw_pile, discard_pile = stepEnv(env, draw_pile, discard_pile, action)

        winner = checkEnded(env)
        if winner != -1 or turn > 300:
            break
    env[89] = 0
    for pIdx in range(6):
        env[77] = pIdx
        env[78:83] = nopeTurn(pIdx)
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(
                getAgentState(env, draw_pile, discard_pile), per_player
            )
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
        elif pIdOrder[pIdx] == 4:
            action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)
        elif pIdOrder[pIdx] == 5:
            action, per5 = p5(getAgentState(env, draw_pile, discard_pile), per5)

    win = False
    if np.where(pIdOrder == -1)[0][0] == checkEnded(env):
        win = True
    else:
        win = False

    return win, per_player


@njit()
def n_game_numba(
    p0,
    num_game,
    per_player,
    list_other,
    per1,
    per2,
    per3,
    per4,
    per5,
    p1,
    p2,
    p3,
    p4,
    p5,
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_numba(
            p0, list_other, per_player, per1, per2, per3, per4, per5, p1, p2, p3, p4, p5
        )
        win += winner
    return win, per_player


import importlib.util
import json, sys

try:
    from env import SHORT_PATH
except:
    pass


def load_module_player(player, game_name=None):
    if game_name == None:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/{player}/Agent_player.py"
        )
    else:
        spec = importlib.util.spec_from_file_location(
            "Agent_player", f"{SHORT_PATH}src/Agent/ifelse/{game_name}/{player}.py"
        )
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    return module


@njit()
def random_Env(p_state, per):
    arr_action = getValidActions(p_state)
    arr_action = np.where(arr_action == 1)[0]
    act_idx = np.random.randint(0, len(arr_action))
    return arr_action[act_idx], per


def load_agent(level, *args):
    num_bot = getAgentSize() - 1

    if "_level_" not in globals():
        global _level_
        _level_ = level
        init = True
    else:
        if _level_ != level:
            _level_ = level
            init = True
        else:
            init = False

    if init:
        global _list_per_level_
        global _list_bot_level_
        _list_per_level_ = []
        _list_bot_level_ = []

        if _level_ == 0:
            _list_per_level_ = [
                np.array([[0.0]], dtype=np.float64) for _ in range(num_bot)
            ]
            _list_bot_level_ = [bot_lv0 for _ in range(num_bot)]
        else:
            env_name = sys.argv[1]
            if len(args) > 0:
                dict_level = json.load(
                    open(f"{SHORT_PATH}src/Log/check_system_about_level.json")
                )
            else:
                dict_level = json.load(open(f"{SHORT_PATH}src/Log/level_game.json"))

            if str(_level_) not in dict_level[env_name]:
                raise Exception("Hiện tại không có level này")

            lst_agent_level = dict_level[env_name][str(level)][2]
            lst_module_level = [
                load_module_player(lst_agent_level[i]) for i in range(num_bot)
            ]
            for i in range(num_bot):
                data_agent_level = np.load(
                    f"{SHORT_PATH}src/Agent/{lst_agent_level[i]}/Data/{env_name}_{level}/Train.npy",
                    allow_pickle=True,
                )
                _list_per_level_.append(
                    lst_module_level[i].convert_to_test(data_agent_level)
                )
                _list_bot_level_.append(lst_module_level[i].Test)

    return _list_bot_level_, _list_per_level_


def run(
    p0: any = bot_lv0,
    num_game: int = 100,
    per_player: np.ndarray = np.array([[0.0]]),
    level: int = 0,
    *args,
):
    num_bot = getAgentSize() - 1
    list_other = np.array([-1] + [i + 1 for i in range(num_bot)])
    try:
        check_njit = check_run_under_njit(p0)
    except:
        check_njit = False

    load_agent(level, *args)

    if check_njit:
        return n_game_numba(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_per_level_[4],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
            _list_bot_level_[4],
        )
    else:
        return n_game_normal(
            p0,
            num_game,
            per_player,
            list_other,
            _list_per_level_[0],
            _list_per_level_[1],
            _list_per_level_[2],
            _list_per_level_[3],
            _list_per_level_[4],
            _list_bot_level_[0],
            _list_bot_level_[1],
            _list_bot_level_[2],
            _list_bot_level_[3],
            _list_bot_level_[4],
        )


def one_game_normal(
    p0, pIdOrder, per_player, per1, per2, per3, per4, per5, p1, p2, p3, p4, p5
):
    env, draw_pile, discard_pile = initEnv()

    winner = -1
    turn = 0
    while True:
        turn += 1
        phase = env[89]
        main_id = env[77]
        nope_id = env[95]
        last_action = env[94]
        if phase == 0:
            pIdx = int(main_id)
        elif phase == 1:
            pIdx = int(nope_id)
        elif phase == 2:
            pIdx = int(main_id)
        elif phase == 3:
            if last_action == 3:
                pIdx = int(env[96])
            else:
                pIdx = int(main_id)
        elif phase == 4:
            pIdx = int(main_id)
        elif phase == 5:
            pIdx = int(main_id)

        if pIdOrder[pIdx] == -1:
            action, per_player = p0(
                getAgentState(env, draw_pile, discard_pile), per_player
            )
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
        elif pIdOrder[pIdx] == 4:
            action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)
        elif pIdOrder[pIdx] == 5:
            action, per5 = p5(getAgentState(env, draw_pile, discard_pile), per5)
        #  print(f'Nope count: {env[76]}')
        #  print(f'Player {pIdx} Valid Action: {np.where(getValidActions(getAgentState(env,draw_pile,discard_pile))==1)[0]}')
        #  print(f'Turn {turn}, Phase {int(phase)} Player {pIdx} action {action} List player {env[83:89]}' )
        #  print(f'See the future: {visualCard(env[91:94])}')
        #  print()
        env, draw_pile, discard_pile = stepEnv(env, draw_pile, discard_pile, action)
        #  print(f'Draw pile: {visualCard(draw_pile)}')
        #  print("____________________________________________________")
        #  print()
        winner = checkEnded(env)
        if winner != -1 or turn > 300:
            #  print(f'Winner: {winner}')
            break
    env[89] = 0
    for pIdx in range(6):
        env[77] = pIdx
        env[78:83] = nopeTurn(pIdx)
        if pIdOrder[pIdx] == -1:
            action, per_player = p0(
                getAgentState(env, draw_pile, discard_pile), per_player
            )
        elif pIdOrder[pIdx] == 1:
            action, per1 = p1(getAgentState(env, draw_pile, discard_pile), per1)
        elif pIdOrder[pIdx] == 2:
            action, per2 = p2(getAgentState(env, draw_pile, discard_pile), per2)
        elif pIdOrder[pIdx] == 3:
            action, per3 = p3(getAgentState(env, draw_pile, discard_pile), per3)
        elif pIdOrder[pIdx] == 4:
            action, per4 = p4(getAgentState(env, draw_pile, discard_pile), per4)
        elif pIdOrder[pIdx] == 5:
            action, per5 = p5(getAgentState(env, draw_pile, discard_pile), per5)

    win = False
    if np.where(pIdOrder == -1)[0][0] == checkEnded(env):
        win = True
    else:
        win = False

    return win, per_player


def n_game_normal(
    p0,
    num_game,
    per_player,
    list_other,
    per1,
    per2,
    per3,
    per4,
    per5,
    p1,
    p2,
    p3,
    p4,
    p5,
):
    win = 0
    for _n in range(num_game):
        np.random.shuffle(list_other)
        winner, per_player = one_game_normal(
            p0, list_other, per_player, per1, per2, per3, per4, per5, p1, p2, p3, p4, p5
        )
        win += winner
    return win, per_player


@njit()
def check_run_under_njit(Agent):
    return True


#  print(n_game_normal(bot_lv0, 1, np.array([0.]), np.array([-1] + [i+1 for i in range(5)]),
#                                  np.array([0.]), np.array([0.]), np.array([0.]),np.array([0.]),np.array([0.]),
#                                  bot_lv0, bot_lv0, bot_lv0,bot_lv0,bot_lv0)[0])
